# -*- coding: utf-8 -*-
#
# This module defines the SkyModel object used by LSMTool for all operations.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import logging
from . import _logging
from . import tableio
from . import operations
from .operations_lib import normalize_ra, normalize_dec

# Python 3 compatibility
try:
    dict.iteritems
except AttributeError:
    # Python 3
    def itervalues(d):
        return iter(d.values())

    def iteritems(d):
        return iter(d.items())
    numpy_type = "U"
else:
    # Python 2
    def itervalues(d):
        return d.itervalues()

    def iteritems(d):
        return d.iteritems()
    numpy_type = "S"


class SkyModel(object):
    """
    Object that stores the sky model and provides methods for accessing it.
    """
    def __init__(self, fileName, beamMS=None, checkDup=False, VOPosition=None,
                 VORadius=None):
        """
        Initializes SkyModel object.

        Parameters
        ----------
        fileName : str
            Input ASCII file from which the sky model is read (must respect the
            makesourcedb format), name of VO service to query (must be one of
            'WENSS', 'NVSS', 'TGSS', or 'GSM'), or dict (single source only)
        beamMS : str, optional
            Measurement set from which the primary beam will be estimated. A
            column of attenuated Stokes I fluxes will be added to the table
        checkDup: bool, optional
            If True, the sky model is checked for duplicate sources (with the
            same name)
        VOPosition : list of floats
            A list specifying a new position as [RA, Dec] in either makesourcedb
            format (e.g., ['12:23:43.21', '+22.34.21.2']) or in degrees (e.g.,
            [123.2312, 23.3422]) for a cone search
        VORadius : float or str, optional
            Radius in degrees (if float) or 'value unit' (if str; e.g.,
            '30 arcsec') for cone search region in degrees

        Examples
        --------
        Create a SkyModel object::

            >>> s = SkyModel('sky.model')

        Create a SkyModel object with a beam MS so that apparent fluxes will
        be available::

            >>> s = SkyModel('sky.model', beamMS='SB100.MS')

        Load a WENSS catalog into a SkyModel object::

            >>> s = SkyModel('WENSS', VOPosition=[212.8352792, 52.202644],
                VOradius=5.0)

        """
        from astropy.table import Table
        from .tableio import processFormatString, processLine, createTable

        self.log = logging.getLogger('LSMTool')
        self.history = []
        if type(fileName) is str:
            if fileName.lower() in tableio.allowedVOServices:
                self.log.debug("Attempting to load model from VO service '{0}'...".format(fileName))
                self.table = tableio.coneSearch(fileName, VOPosition, VORadius)
                self.log.debug("Successfully loaded model from VO service '{0}'".format(fileName))
                self._fileName = None
                self._addHistory("LOAD (from {0} at position {1})".format(fileName, VOPosition))
            elif fileName.lower() == 'tgss':
                self.log.debug("Attempting to load model from TGSS...")
                fileObj = tableio.getTGSS(VOPosition, VORadius)
                self.table = Table.read(fileObj.name, format='makesourcedb')
                fileObj.close()
                self.log.debug("Successfully loaded model from TGSS")
                self._fileName = None
                self._addHistory("LOAD (from TGSS at position {0})".format(VOPosition))
            elif fileName.lower() == 'gsm':
                self.log.debug("Attempting to load model from GSM...")
                fileObj = tableio.getGSM(VOPosition, VORadius)
                self.table = Table.read(fileObj.name, format='makesourcedb')
                fileObj.close()
                self.log.debug("Successfully loaded model from GSM")
                self._fileName = None
                self._addHistory("LOAD (from GSM at position {0})".format(VOPosition))
            elif fileName.lower() == 'lotss':
                self.log.debug("Attempting to load model from LoTSS...")
                self.table = tableio.getLoTSS(VOPosition, VORadius)
                self.log.debug("Successfully loaded model from LoTSS")
                self._fileName = None
                self._addHistory("LOAD (from LoTSS at position {0})".format(VOPosition))
            else:
                self.log.debug("Attempting to load model from file '{0}'...".format(fileName))
                self.table = Table.read(fileName, format='makesourcedb')
                self.log.debug("Successfully loaded model from file '{0}'".format(fileName))
                self._fileName = fileName
                self._addHistory("LOAD (from file '{0}')".format(fileName))
        elif type(fileName) is dict:
            self.log.debug("Attempting to create model from input dict...")
            # Create header
            formatString = '#FORMAT = ' + ', '.join(fileName.keys())

            # Process the header
            colNames, hasPatches, colDefaults, metaDict = processFormatString(formatString)

            # Process the model
            outlines = []
            stringValues = ['{0}'.format(v) for v in fileName.values()]
            line = ', '.join(stringValues)
            outline, metaDict = processLine(line, metaDict, colNames)
            if outline is not None:
                outlines.append(outline)
            outlines.append('\n')  # needed in case of single-line sky models
            table = createTable(outlines, metaDict, colNames, colDefaults)
            self.table = table
            self.log.debug("Successfully created model from input dict")
            self._fileName = None
            self._addHistory("LOAD (from input dict)")
        else:
            raise ValueError("Filename not understood. Exiting...")

        if beamMS is not None:
            self.beamMS = beamMS
            self._hasBeam = True
            self.beamTime = 0.5
        else:
            self.beamMS = None
            self._hasBeam = False
            self.beamTime = None

        if checkDup:
            self.log.debug('Checking model for duplicate lines...')
            self._clean()

        self.log.debug('Processing patches (if any)...')
        self._patchMethod = None
        self._updateGroups()

        self.log.debug("Successfully loaded sky model")

    def __len__(self):
        """
        Returns the table len() value (number of rows).
        """
        return self.table.__len__()

    def __str__(self):
        """
        Returns string with info about sky model contents.
        """
        return self.table.__str__()

    def _updateGroups(self):
        """
        Updates the grouping of the table by patch name.
        """
        if 'Patch' in self.table.keys():
            self.table = self.table.group_by('Patch')
            self.hasPatches = True
        else:
            self.hasPatches = False

    def _addHistory(self, entry=""):
        """
        Adds entry to the history with current date and time

        Parameters
        ----------
        entry : str, optional
            String to add to history

        """
        import datetime
        current_time = str(datetime.datetime.now()).split('.')[0]
        self.history.append(current_time + ": " + str(entry))

    def _info(self, useLogInfo=False):
        """
        Prints information about the sky model.
        """
        import numpy as np

        if self.hasPatches:
            nPatches = len(set(self.getPatchNames()))
        else:
            nPatches = 0

        nPoint = len(np.where(self.getColValues('Type') == 'POINT')[0])
        nGaus = len(np.where(self.getColValues('Type') == 'GAUSSIAN')[0])

        if nPatches == 1:
            plur = ''
        else:
            plur = 'es'
        if useLogInfo:
            logCall = self.log.info
        else:
            logCall = self.log.debug

        x, y, refRA, refDec = self._getXY()
        totFlux = np.sum(self.getColValues('I', units='Jy'))

        info = 'Model contains {0} sources in {1} patch{2} of which:\n'\
               '      {3} are type POINT\n'\
               '      {4} are type GAUSSIAN\n'\
               '      Associated beam MS: {5}\n'\
               '      Approximate RA, Dec of center: {6}, {7}\n'\
               '      Total flux: {8} Jy\n\n'\
               '      History:\n'\
               '      {9}'.format(len(self.table), nPatches, plur,
                                  nPoint, nGaus, self.beamMS, refRA, refDec, totFlux,
                                  '\n      '.join(self.history))
        logCall(info)
        return info

    def info(self):
        """
        Prints information about the sky model.
        """
        infoStr = self._info(useLogInfo=True)

    def copy(self):
        """
        Returns a copy of the sky model.
        """
        import copy

        # The logger's stream handlers are not copyable with deepcopy, so copy
        # them by hand:
        self.log = None
        LSMCopy = copy.deepcopy(self)
        LSMCopy.log = logging.getLogger('LSMTool')
        LSMCopy._addHistory('COPY')
        self.log = logging.getLogger('LSMTool')

        return LSMCopy

    def more(self, colName=None, patchName=None, sourceName=None, sortBy=None,
             lowToHigh=False):
        """
        Prints the sky model table to the screen with more-like commands.

        Parameters
        ----------
        colName : str, list of str, optional
            Name of column or columns to print. If None, all columns are printed
        patchName : str, list of str, optional
            If given, returns column values for specified patch or patches only
        sourceName : str, list of str, optional
            If given, returns column value for specified source or sources only
        sortBy : str or list of str, optional
            Name of columns to sort on. If None, no sorting is done. If
            a list is given, sorting is done on the columns in the order given
        lowToHigh : bool, optional
            If True, sort values from low to high instead of high to low

        Examples
        --------
        Print the entire model::

            >>> s.more()

        Print only the 'Name' and 'I' columns for the 'bin0' patch::

            >>> s.more(['Name', 'I'], 'bin0', sortBy=['I'])

        """
        if patchName is not None and sourceName is not None:
            raise ValueError('patchName and sourceName cannot both be specified.')

        table = self.table

        # Get columns
        colName = self._verifyColName(colName)
        if colName is not None:
            if type(colName) is str:
                colName = [colName]  # needed in order to get a table instead of a column
            table = table[colName]

        # Get patches
        if patchName is not None:
            pindx = self._getNameIndx(patchName, patch=True)
            if pindx is not None:
                table = table.groups[pindx]

        # Get sources
        if sourceName is not None:
            sindx = self._getNameIndx(sourceName)
            if sindx is not None:
                table = table[sindx]

        # Sort if desired
        if sortBy is not None:
            colName = self._verifyColName(sortBy)
            indx = table.argsort(colName)
            if not lowToHigh:
                indx = indx[::-1]
            table = table[indx]

        table.more(show_unit=True)

    def _verifyColName(self, colName, onlyExisting=True, applyBeam=False,
                       quiet=False):
        """
        Verifies that column(s) exist and returns correctly formatted string or
        list of strings suitable for accessing the data table.

        Parameters
        ----------
        colName : str, list of str
            Name of column or columns
        onlyExisting : bool, optional
            If True, only columns that exist in the table are allowed. If False,
            columns that are valid but do not exist in the table are returned
            without error.
        applyBeam : bool, optional
            If True and colName = 'I', the attenuated Stokes I column will be
            returned
        quiet : bool, optional
            If True, errors will be suppressed

        Returns
        -------
        colName : str, None
            Properly formatted name of column or None if colName not found

        """
        if type(colName) is str:
            colNameLower = colName.lower()
            if colNameLower not in tableio.allowedColumnNames:
                if not quiet:
                    raise ValueError('Column name "{0}" is not a valid makesourcedb '
                                     'column.'.format(colName))
                return None
            else:
                colNameKey = tableio.allowedColumnNames[colNameLower]
            if colNameKey not in self.table.keys() and onlyExisting:
                if not quiet:
                    raise ValueError('Column name "{0}" not found in sky model.'.
                                     format(colName))
                return None

        elif type(colName) is list:
            colNameLower = [c.lower() for c in colName]
            for name in colNameLower[:]:
                badNames = []
                if name not in tableio.allowedColumnNames:
                    badNames.append(name)
                    colNameLower.remove(name)
                else:
                    colNameKey = tableio.allowedColumnNames[name]
                    if colNameKey not in self.table.keys():
                        badNames.append(name)
                        colNameLower.remove(name)

            if len(badNames) > 0:
                if len(badNames) == 1:
                    plur = ''
                else:
                    plur = 's'
                if not quiet:
                    self.log.warn("Column name{0} '{1}' not recognized. Ignoring.".
                                  format(plur, ','.join(badNames)))
            if len(colNameLower) == 0:
                return None
            else:
                colNameKey = [tableio.allowedColumnNames[n] for n in colNameLower]
        else:
            colNameKey = None

        return colNameKey

    def getPatchPositions(self, patchName=None, asArray=False, method=None,
                          applyBeam=False, perPatchProjection=True):
        """
        Returns arrays or a dict of patch positions (as {'patchName':(RA, Dec)}).

        Parameters
        ----------
        patchName : str or list, optional
            List of patch names for which the positions are desired
        asArray : bool, optional
            If True, returns arrays of RA, Dec instead of a dict
        method : None or str, optional
            This parameter specifies the method used to calculate the patch
            positions. If None, the current patch positions stored in the sky
            model, if any, will be returned.
            - 'mid' => calculate the midpoint of the patch
            - 'mean' => calculate the mean RA and Dec of the patch
            - 'wmean' => calculate the flux-weighted mean RA and Dec of the patch
            - None => current patch positions are returned
            Note that the mid, mean, and wmean positions are calculated from TAN-
            projected values.
        applyBeam : bool, optional
            If True, fluxes used as weights will be attenuated by the beam.
        perPatchProjection : bool, optional
            If True, a different projection center is used per patch. If False,
            a single projection center is used for all patches.

        Returns
        -------
        positions : numpy array or dict
            (RA, Dec) arrays (if asArray is False) of patch positions or a
            dictionary of {'patchName':(RA, Dec)}.

        Examples
        --------
        Get the current patch positions::

            >>> s.getPatchPositions()
            {'bin0': [<Angle 91.77565208333331 deg>, <Angle 41.57834805555555 deg>],
             'bin1': [<Angle 91.59991874999997 deg>, <Angle 41.90387583333333 deg>],
             'bin2': [<Angle 90.83773333333332 deg>, <Angle 42.189861944444445 deg>],

        Get them as RA and Dec arrays in degrees::

            >>> s.getPatchPositions(asArray=True)
            (array([ 91.77565208,  91.59991875,  90.83773333]),
             array([ 41.57834806,  41.90387583,  42.18986194]))

        Calculate the flux-weighted mean positions of each patch::

            >>> s.getPatchPositions(method='wmean', asArray=True)

        """
        import numpy as np
        from .operations_lib import radec2xy, xy2radec
        from astropy.table import Column
        from .tableio import RA2Angle, Dec2Angle

        if self.hasPatches:
            if patchName is None:
                patchName = self.getPatchNames()
            if type(patchName) is str or type(patchName) is np.string_ or type(patchName) is np.unicode_:
                patchName = [patchName]
            if method is None:
                patchDict = {}
                for patch in patchName:
                    if patch in self.table.meta:
                        patchDict[patch] = self.table.meta[patch]
                    else:
                        patchDict[patch] = [RA2Angle(0.0)[0], Dec2Angle(0.0)[0]]
            else:
                patchDict = {}

                # Add projected x and y columns.
                if perPatchProjection:
                    # Each patch has a different projection center
                    xAll = []  # has length = num of sources
                    yAll = []
                    midRAAll = []  # has length = num of patches
                    midDecAll = []
                    for name in patchName:
                        x, y, midRA, midDec = self._getXY(patchName=name)
                        xAll.extend(x)
                        yAll.extend(y)
                        midRAAll.append(midRA)
                        midDecAll.append(midDec)
                else:
                    xAll, yAll, midRA, midDec = self._getXY()
                    midRAAll = []  # has length = num of patches
                    midDecAll = []
                    for name in patchName:
                        midRAAll.append(midRA)
                        midDecAll.append(midDec)

                xCol = Column(name='X', data=xAll)
                yCol = Column(name='Y', data=yAll)
                self.table.add_column(xCol)
                self.table.add_column(yCol)

                if method == 'mid':
                    minX = self._getMinColumn('X')
                    maxX = self._getMaxColumn('X')
                    minY = self._getMinColumn('Y')
                    maxY = self._getMaxColumn('Y')
                    midX = minX + (maxX - minX) / 2.0
                    midY = minY + (maxY - minY) / 2.0
                    for i, name in enumerate(patchName):
                        gRA = RA2Angle(xy2radec([midX[i]], [midY[i]], midRAAll[i],
                                       midDecAll[i])[0])[0]
                        gDec = Dec2Angle(xy2radec([midX[i]], [midY[i]], midRAAll[i],
                                         midDecAll[i])[1])[0]
                        patchDict[name] = [gRA, gDec]
                elif method == 'mean' or method == 'wmean':
                    if method == 'mean':
                        weight = False
                    else:
                        weight = True
                    meanX = self._getAveragedColumn('X', applyBeam=applyBeam,
                                                    weight=weight)
                    meanY = self._getAveragedColumn('Y', applyBeam=applyBeam,
                                                    weight=weight)
                    for i, name in enumerate(patchName):
                        gRA = RA2Angle(xy2radec([meanX[i]], [meanY[i]], midRAAll[i],
                                       midDecAll[i])[0])[0]
                        gDec = Dec2Angle(xy2radec([meanX[i]], [meanY[i]], midRAAll[i],
                                         midDecAll[i])[1])[0]
                        patchDict[name] = [gRA, gDec]
                self.table.remove_column('X')
                self.table.remove_column('Y')

            if asArray:
                RA = []
                Dec = []
                for patch in patchName:
                    RA.append(patchDict[patch][0].value)
                    Dec.append(patchDict[patch][1].value)
                return np.array(RA), np.array(Dec)
            else:
                return patchDict

        else:
            return None

    def setPatchPositions(self, patchDict=None, method='mid', applyBeam=False,
                          perPatchProjection=True):
        """
        Sets the patch positions.

        Parameters
        ----------
        patchDict : dict, optional
            Dict specifying patch names and positions as {'patchName':[RA, Dec]}
            where both RA and Dec are degrees J2000 or in makesourcedb format.
            If None, positions are set for all patches using the method given
            by the 'method' parameter.
        method : None or str, optional
            If no patchDict is given, this parameter specifies the method used
            to set the patch positions:
            - 'mid' => the position is set to the midpoint of the patch
            - 'mean' => the position is set to the mean RA and Dec of the patch
            - 'wmean' => the position is set to the flux-weighted mean RA and
            Dec of the patch
            - 'zero' => set all positions to [0.0, 0.0]

            Note that the mid, mean, and wmean positions are calculated from TAN-
            projected values.
        applyBeam : bool, optional
            If True, fluxes used as weights will be attenuated by the beam.
        perPatchProjection : bool, optional
            If True, a different projection center is used per patch. If False,
            a single projection center is used for all patches.

        Examples
        --------
        Set all patch positions to their (projected) midpoints::

            >>> s.setPatchPositions()

        Set all patch positions to their (projected) flux-weighted mean
        positions::

             >>> s.setPatchPositions(method='wmean')

        Set new position for the 'bin0' patch only::

            >>> s.setPatchPositions({'bin0': [123.231, 23.4321]})

        """
        from .tableio import RA2Angle, Dec2Angle

        if self.hasPatches:
            if method not in ['mid', 'mean', 'wmean', 'zero']:
                raise ValueError('Invalid method parameter')

            if patchDict is None:
                # Delete any previous patch positions
                patchNames = self.getPatchNames()
                for patchName in patchNames:
                    if patchName in self.table.meta:
                        self.table.meta.pop(patchName)
                if method == 'zero':
                    patchDict = {}
                    for n in patchNames:
                        patchDict[n] = [RA2Angle(0.0), Dec2Angle(0.0)]
                else:
                    patchDict = self.getPatchPositions(method=method, applyBeam=applyBeam,
                                                       perPatchProjection=perPatchProjection)

            for patch, pos in iteritems(patchDict):
                if type(pos[0]) is str or type(pos[0]) is float:
                    pos[0] = RA2Angle(pos[0])
                if type(pos[1]) is str or type(pos[1]) is float:
                    pos[1] = Dec2Angle(pos[1])
                self.table.meta[patch] = pos
            self._addHistory("SETPATCHPOSITIONS (method = '{0}')".format(method))
        else:
            raise RuntimeError('Sky model does not have patches.')

    def _getXY(self, patchName=None, crdelt=None, byPatch=False):
        """
        Returns lists of projected x and y values for all sources.

        Parameters
        ----------
        patchName : str, optional
            If given, return x and y for specified patch only
        crdelt: float, optional
            Delta in degrees for sky grid
        byPatch : bool, optional
            Use patches instead of by sources

        Returns
        -------
        x, y, midRA, midDec : numpy array, numpy array, float, float
            arrays of x and y values and the midpoint RA and
            Dec values

        """
        from .operations_lib import radec2xy, xy2radec
        import numpy as np

        if len(self.table) == 0:
            return [0], [0], 0, 0

        if byPatch:
            if 'Patch' not in self.table.keys():
                raise ValueError('Sky model must be grouped before "byPatch" can be used.')
            RA = self.getColValues('Ra', aggregate='wmean')
            Dec = self.getColValues('Dec', aggregate='wmean')
        else:
            RA = self.getColValues('Ra')
            Dec = self.getColValues('Dec')
            if patchName is not None:
                ind = self.getRowIndex(patchName)
                RA = RA[ind]
                Dec = Dec[ind]
        x, y = radec2xy(RA, Dec, crdelt=crdelt)

        # Refine x and y using midpoint
        if len(x) > 1:
            xmid = min(x) + (max(x) - min(x)) / 2.0
            ymid = min(y) + (max(y) - min(y)) / 2.0
            xind = np.argsort(x)
            yind = np.argsort(y)
            try:
                midxind = np.where(np.array(x)[xind] > xmid)[0][0]
                midyind = np.where(np.array(y)[yind] > ymid)[0][0]
                midRA = RA[xind[midxind]]
                midDec = Dec[yind[midyind]]
                x, y = radec2xy(RA, Dec, midRA, midDec, crdelt=crdelt)
            except IndexError:
                midRA = RA[0]
                midDec = Dec[0]
        else:
            midRA = RA[0]
            midDec = Dec[0]

        return np.array(x), np.array(y), normalize_ra(midRA), normalize_dec(midDec)

    def getDefaultValues(self):
        """
        Returns dict of {colName:default} values for all columns with defaults.

        Returns
        -------
        defaultDict : dict
            Dict of {colName:default} values

        """
        colNames = self.getColNames()
        defaultDict = {}
        for colName in colNames:
            if colName in self.table.meta:
                defaultDict[colName] = self.table.meta[colName]
        return defaultDict

    def setDefaultValues(self, colDict):
        """
        Sets default column values.

        Parameters
        ----------
        colDict : dict
            Dict specifying column names and default values as
            {'colName':value} where the value is in the units accepted by
            makesourcedb (e.g., Hz for 'ReferenceFrequency').

        Examples
        --------
        Set new default value for ReferenceFrequency::

            >>> s.setDefaultValues({'ReferenceFrequency': 140e6})

        """
        for colName, default in iteritems(colDict):
            self.table.meta[colName] = default

    def ungroup(self):
        """
        Removes all patches from the sky model.

        Examples
        --------
        Remove all patches::

            >>> s.ungroup()

        """
        if self.hasPatches:
            for patchName in self.getPatchNames():
                if patchName in self.table.meta:
                    self.table.meta.pop(patchName)
            self.table.remove_column('Patch')
            self._updateGroups()
            self._addHistory('UNGROUP')
            self._info()

    def getColNames(self):
        """
        Returns a list of all available column names.

        Returns
        -------
        colNames : list
            List of all column names

        Examples
        --------
        Get column names::

            >>> s.getColNames()

        """
        return self.table.keys()

    def getColValues(self, colName, units=None, aggregate=None,
                     applyBeam=False):
        """
        Returns a numpy array of column values.

        Parameters
        ----------
        colName : str
            Name of column
        units : str, optional
            Output units (the values are converted as needed). By default, the
            units are those used by makesourcedb, with the exception of RA and
            Dec which have default output units of degrees.
        aggregate : {'sum', 'mean', 'wmean', 'min', max'}, optional
            If set, the array returned will be of values aggregated
            over the patch members. The following aggregation functions are
            available:
                - 'sum': sum of patch values
                - 'mean': mean of patch values
                - 'wmean': Stokes-I-weighted mean of patch values
                - 'min': minimum of patch values
                - 'max': maximum of patch values
            Note that, in some cases, certain aggregation functions will not
            produce meaningful results. For example, asking for the sum of
            the MajorAxis values per patch will not give a good indication of
            the size of the patch (to get the sizes, use the getPatchSizes()
            method). Additionally, applying the 'mean' or 'wmean' functions to
            the RA or Dec columns may give strange results near the poles or
            near RA = 0h. For aggregated RA and Dec values, use the
            getPatchPositions() method instead which projects the sources onto
            the image plane before aggregation.
        applyBeam : bool, optional
            If True, fluxes will be attenuated by the beam. This attenuation
            also applies to fluxes used in aggregation functions.

        Returns
        -------
        colValues : numpy array
            Array of column values. None is returned if column is not found.

        Examples
        --------
        Get Stokes I fluxes in Jy::

            >>> s.getColValues('I')
            array([ 60.4892,   1.2413,   1.216 , ...,   1.12  ,   1.25  ,   1.16  ])

        Get Stokes I fluxes in mJy::

            >>> s.getColValues('I', units='mJy')
            array([ 60489.2,   1241.3,   1216. , ...,   1120. ,   1250. ,   1160. ])

        Get total Stokes I flux for the patches::

            >>> s.getColValues('I', aggregate='sum')
            array([ 61.7305,   1.216 ,   3.9793, ...,   1.12  ,   1.25  ,   1.16  ])

        Get flux-weighted average RA and Dec for the patches. As noted above, the
        getColValues() method is not appropriate for use with RA or Dec, so
        we must use getPatchPositions() instead::

            >>> RA, Dec = s.getPatchPositions(method='wmean', asArray=True)

        """
        colName = self._verifyColName(colName)
        if colName is None:
            return None
        if type(colName) is list:
            if len(colName) > 1:
                raise ValueError('Only one column can be specified.')
            else:
                colName = colName[0]

        allowedFcns = ['sum', 'mean', 'wmean', 'min', 'max']
        if aggregate not in allowedFcns and aggregate is not None:
            raise ValueError("Value of parameter 'aggregate' not understood.")
        if aggregate in allowedFcns and self.hasPatches:
            col = self._getAggregatedColumn(colName, aggregate, applyBeam=applyBeam)
        else:
            col = self._getColumn(colName, applyBeam=applyBeam)

        if col is None:
            return None

        if hasattr(col, 'filled'):
            outcol = col.filled().copy()
        else:
            outcol = col.copy()

        if units is not None:
            outcol.convert_unit_to(units)

        return outcol.data

    def setColValues(self, colName, values, mask=None, index=None):
        """
        Sets column values.

        Parameters
        ----------
        colName : str
            Name of column. If not already present in the table, a new column
            will be created.
        values : list, numpy array, or dict
            Array of values or dict of {sourceName:value} pairs. If list or
            array, the length must match the number of rows in the table. If
            dict, missing values will be masked unless already present. Values
            are assumed to be in units required by makesourcedb.
        mask : list or array of bools, optional
            If values is a list or array, a mask can be specified (True means
            the value is masked).
        index : int, optional
            Index that specifies the column position in the table, if column is
            not already present in the table.

        Examples
        --------
        Set Stokes I fluxes::

            >>> s.setColValues('I', [1.0, 1.1, 1.2, 0.0, 1.3], mask=[False,
                    False, False, True, False])

        """
        from astropy.table import Column
        from .tableio import RA2Angle, Dec2Angle
        import numpy as np

        colName = self._verifyColName(colName, onlyExisting=False)
        if colName is None:
            return None
        if type(colName) is list:
            if len(colName) > 1:
                raise ValueError('Only one column can be specified.')
            else:
                colName = colName[0]

        if isinstance(values, dict):
            if colName in self.table.keys():
                data = self.table[colName].data
                mask = self.table[colName].mask
            else:
                data = [0] * len(self.table)
                mask = [True] * len(self.table)
            for sourceName, value in iteritems(values):
                indx = self._getNameIndx(sourceName)
                if colName == 'Ra':
                    val = RA2Angle(value)[0]
                elif colName == 'Dec':
                    val = Dec2Angle(value)[0]
                else:
                    val = value
                data[indx] = val
                mask[indx] = False
        else:
            if len(values) != len(self.table):
                raise ValueError('Length of input values must match length of table.')
            else:
                if colName == 'Ra':
                    vals = RA2Angle(values)
                elif colName == 'Dec':
                    vals = Dec2Angle(values)
                else:
                    vals = values
                data = vals

        if mask is not None:
            data = np.ma.masked_array(data, mask)
        else:
            data = np.array(data)
        if colName in self.table.keys():
            units = self.table.columns[colName].unit
            self.table[colName] = data
            self.table.columns[colName].unit = units
        else:
            if colName == 'Patch':
                # Specify length of 50 characters
                newCol = Column(name=colName, data=data, dtype='{}50'.format(numpy_type))
            else:
                newCol = Column(name=colName, data=data)
            self.table.add_column(newCol, index=index)
        if colName == 'Patch':
            self._updateGroups()

    def getRowValues(self, rowName):
        """
        Returns an astropy table or table row for specified source or patch.

        Parameters
        ----------
        rowName : str
            Name of the source or patch

        Returns
        -------
        rowValues : astropy table or row
            Table (if more than one source) or row (if one source). None is
            returned if source is not found.

        Examples
        --------
        Get row values for the source 'src1'::

            >>> rows = s.getRowValues('src1')

        Sum over the fluxes of sources in the 'bin1' patch::

            >>> tot = 0.0
            >>> for row in s.getRowValues('bin1'): tot += row['I']

        """
        sourceNames = self.getColValues('Name')
        patchNames = self.getPatchNames()
        if patchNames is None:
            patchNames = []
        if rowName in sourceNames:
            indx = self._getNameIndx(rowName)
            return self.table.filled()[indx]
        elif rowName in patchNames:
            pindx = self._getNameIndx(rowName, patch=True)
            table = self.table.groups[pindx]
            table = table.group_by('Patch')  # ensure that grouping is preserved
            return table
        else:
            raise ValueError("Row name '{0}' not recognized.".format(rowName))

    def getRowIndex(self, rowName):
        """
        Returns index or indices for specified source or patch as a list.

        Parameters
        ----------
        rowName : str
            Name of the source or patch

        Returns
        -------
        indices : list
            List of indices. ValueError is raised if the source is not found.

        Examples
        --------
        Get row index for the source 'src1'::

            >>> s.getRowIndex('src1')
            [0]

        Get row indices for the patch 'bin1' and verify the patch name::

            >>> ind = s.getRowIndex('bin1')
            >>> print(s.getColValues('patch')[ind])
            ['bin1', 'bin1', 'bin1']

        """
        import numpy as np

        sourceNames = self.getColValues('Name')
        if self.hasPatches:
            patchNames = self.getColValues('Patch')
        else:
            patchNames = []

        if rowName in sourceNames:
            return self._getNameIndx(rowName)
        elif rowName in patchNames:
            return np.where(patchNames == rowName)[0].tolist()
        else:
            raise ValueError("Row name '{0}' not recognized.".format(rowName))

    def setRowValues(self, values, mask=None, returnVerified=False):
        """
        Sets values for a single row.

        If a row with the given name already exists, its values are
        updated. If not, a new row is made and appended to the table.

        Parameters
        ----------
        values : list, numpy array, or dict
            Array of values or dict of {colName:value} pairs. If list or
            array, the length must match the number and order of the columns in
            the table. If dict, missing values will be masked unless already
            present.
        mask : list or array of bools, optional
            If values is a list or array, a mask can be specified (True means
            the value is masked).
        returnVerified : bool, optional
            If True, the values are verified and returned, allowing them to be
            passed to table.add_row().

        Examples
        --------
        Set row values for the source 'src1' (which can be a new source or an
        existing source)::

            >>> s.setRowValues({'Name':'src1', 'Ra':213.123, 'Dec':23.1232,
                'I':23.2, 'Type':'POINT'})

        The RA and Dec values can be in degrees (as above) or in makesourcedb
        format. E.g.::

            >>> s.setRowValues({'Name':'src1', 'Ra':'12:22:21.1',
                'Dec':'+14.46.31.5', 'I':23.2, 'Type':'POINT'})

        """
        # Read model into astropy table object
        tempLSM = SkyModel(values)

        # Concatenate tables
        self.concatenate(tempLSM, matchBy='name', keep='from2', inheritPatches=False)

    def getPatchSizes(self, units=None, weight=False, applyBeam=False):
        """
        Returns array of patch sizes.

        Parameters
        ----------
        units : str, optional
            Units for returned sizes (e.g., 'arcsec', 'degree')
        weight : bool, optional
            If True, weight the source positions inside the patch by flux
        applyBeam : bool, optional
            If True and weight is True, attenuate the fluxes used for weighting
            by the beam

        Returns
        -------
        data : numpy array
            Array of patch sizes. None is returned if the sky model
            does not have patches

        """
        if self.hasPatches:
            col = self._getSizeColumn(weight=weight, applyBeam=applyBeam)
            if units is not None:
                col.convert_unit_to(units)
            return col.data
        else:
            return None

    def getPatchNames(self):
        """
        Returns array of all patch names in the sky model, with duplicates removed.

        Note: use getColValues('Patch') if you want the patch names for each source
        in the sky model.

        Returns
        -------
        names : numpy array
            Array of patch names. None is returned if the sky model
            does not have patches

        """
        if self.hasPatches:
            col = self.table.groups.keys['Patch']
            if hasattr(col, 'filled'):
                outcol = col.filled().copy()
            else:
                outcol = col.copy()
            return outcol.data
        else:
            return None

    def _getNameIndx(self, name, patch=False):
        """
        Returns a list of indices corresponding to the given names.

        Parameters
        ----------
        name : str, list of str
            source or patch name or list of names (UNIX-style wildcards are
            allowed)
        patch : bool
            if True, return the index of the group corresponding to the given
            name; otherwise return the index of the source

        Returns
        -------
        indices : list
            List of indices

        """
        import numpy as np
        import fnmatch

        if patch:
            if self.hasPatches:
                names = self.getPatchNames().tolist()
            else:
                return None
        else:
            names = self.getColValues('Name').tolist()

        if type(name) is str or type(name) is np.string_:
            indx = [i for i, item in enumerate(names) if fnmatch.fnmatch(item, name)]
            if len(indx) == 0:
                return None
            return indx
        elif type(name) is list:
            indx = []
            for n in name:
                badNames = []
                nindx = [i for i, item in enumerate(names) if fnmatch.fnmatch(item, n)]
                if len(nindx) == 0:
                    badNames.append(n)
                else:
                    indx += nindx
            if len(badNames) > 0:
                if len(badNames) == 1:
                    plur = ''
                else:
                    plur = 's'
                self.log.warn("Name{0} '{1}' not recognized. Ignoring.".
                              format(plur, ','.join(badNames)))
            if len(indx) == 0:
                raise ValueError("None of the specified names were found.")
            return indx
        else:
            return None

    def _getColumn(self, colName, applyBeam=False):
        """
        Returns the appropriate column (nonaggregated).

        Parameters
        ----------
        colName : str
            Name of column to get. If not found, None is returned
        applyBeam : bool, optional
            If True, fluxes will be attenuated by the beam

        Returns
        -------
        col : astropy Column
            Nonaggregated Column object

         """
        colName = self._verifyColName(colName)
        if colName is None:
            return None

        col = self.table[colName].copy()

        if applyBeam and colName in ['I', 'Q', 'U', 'V']:
            col = self._applyBeamToCol(col)

        return col

    def _getAggregatedColumn(self, colName, aggregate='sum', applyBeam=False):
        """
        Returns the appropriate column aggregated by group.

        Parameters
        ----------
        colName : str
            Name of column to get. If not found, None is returned
        aggregate : str, optional
            If set, the array returned will be of values aggregated
            over the patch members. The following aggregation functions are
            available:
                - 'sum': sum of patch values
                - 'mean': mean of patch values
                - 'wmean': Stokes I weighted mean of patch values
                - 'min': minimum of patch values
                - 'max': maximum of patch values
        applyBeam : bool, optional
            If True, fluxes will be attenuated by the beam

        Returns
        -------
        col : astropy Column
            Column object with aggregated values

         """
        colName = self._verifyColName(colName)
        if colName is None:
            return None

        if aggregate == 'mean':
            col = self._getAveragedColumn(colName, weight=False,
                                          applyBeam=applyBeam)
        elif aggregate == 'wmean':
            col = self._getAveragedColumn(colName, weight=True,
                                          applyBeam=applyBeam)
        elif aggregate == 'sum':
            col = self._getSummedColumn(colName, applyBeam=applyBeam)
        elif aggregate == 'min':
            col = self._getMinColumn(colName, applyBeam=applyBeam)
        elif aggregate == 'max':
            col = self._getMaxColumn(colName, applyBeam=applyBeam)
        else:
            raise ValueError('Aggregation function not understood.'.format(colName))
        return col

    def _applyBeamToCol(self, col, patch=False):
        """
        Applies beam attenuation to the column values.

        Parameters
        ----------
        col : astropy Column
            Column of flux values to attenuate
        patch : bool, optional
            If True, col is assumed to be aggregated over patches

        Returns
        -------
        col : astropy Column
            Column object with flux values attenuated by the beam

        """
        from .operations_lib import attenuate

        if not self._hasBeam:
            self.log.warn('No beam MS has been specified. No beam attenuation applied.')
            return col

        if patch:
            if self._patchMethod is not None:
                # Try to get patch positions from the meta data
                RADeg, DecDeg = self.getPatchPositions(asArray=True)
            else:
                # If patch positions are not set, use weighted mean positions
                RADeg = self.getColValues('Ra', applyBeam=True, aggregate='wmean')
                DecDeg = self.getColValues('Dec', applyBeam=True, aggregate='wmean')
        else:
            RADeg = self.getColValues('Ra')
            DecDeg = self.getColValues('Dec')

        flux = col.data
        vals = attenuate(self.beamMS, flux, RADeg, DecDeg, timeIndx=self.beamTime)
        col[:] = vals

        return col

    def _getSummedColumn(self, colName, applyBeam=False):
        """
        Returns column summed by group.

        Parameters
        ----------
        colName : str
            Column name
        applyBeam : bool, optional
            If True, fluxes will be attenuated by the beam

        Returns
        -------
        col : astropy Column
            Column object with aggregated sum of values

        """
        import numpy as np

        def npsum(array):
            return np.sum(array, axis=0)

        if hasattr(self.table[colName], 'filled'):
            col = self.table[colName].filled()
            gcol = col.group_by(self.table['Patch'])
            gcol = gcol.groups.aggregate(npsum)
        else:
            gcol = self.table[colName].groups.aggregate(npsum)
        if applyBeam and colName in ['I', 'Q', 'U', 'V']:
            gcol = self._applyBeamToCol(gcol, patch=True)

        return gcol

    def _getMinColumn(self, colName, applyBeam=False):
        """
        Returns column minimum value by group.

        Parameters
        ----------
        colName : str
            Column name
        applyBeam : bool, optional
            If True, fluxes will be attenuated by the beam

        Returns
        -------
        col : astropy Column
            Column object with aggregated min values

        """
        import numpy as np

        def npmin(array):
            return np.min(array, axis=0)

        if hasattr(self.table[colName], 'filled'):
            col = self.table[colName].filled()
            gcol = col.group_by(self.table['Patch'])
            gcol = gcol.groups.aggregate(npmin)
        else:
            gcol = self.table[colName].groups.aggregate(npmin)
        if applyBeam and colName in ['I', 'Q', 'U', 'V']:
            gcol = self._applyBeamToCol(gcol, patch=True)

        return gcol

    def _getMaxColumn(self, colName, applyBeam=False):
        """
        Returns column maximum value by group.

        Parameters
        ----------
        colName : str
            Column name
        applyBeam : bool, optional
            If True, fluxes will be attenuated by the beam

        Returns
        -------
        col : astropy Column
            Column object with aggregated max values

        """
        import numpy as np

        def npmax(array):
            return np.max(array, axis=0)

        if hasattr(self.table[colName], 'filled'):
            col = self.table[colName].filled()
            gcol = col.group_by(self.table['Patch'])
            gcol = gcol.groups.aggregate(npmax)
        else:
            gcol = self.table[colName].groups.aggregate(npmax)
        if applyBeam and colName in ['I', 'Q', 'U', 'V']:
            gcol = self._applyBeamToCol(gcol, patch=True)

        return gcol

    def _getAveragedColumn(self, colName, weight=True, applyBeam=False):
        """
        Returns column averaged by group.

        Parameters
        ----------
        colName : str
            Column name
        weight : bool, optional
            If True, return average weighted by flux
        applyBeam : bool, optional
            If True, fluxes will be attenuated by the beam

        Returns
        -------
        col : astropy Column
            Column object with aggregated mean values

        """
        from astropy.table import Column
        import numpy as np

        if weight:
            def npsum(array):
                return np.sum(array, axis=0)

            if hasattr(self.table[colName], 'filled'):
                vals = self.table[colName].filled().data
            else:
                vals = self.table[colName].data
            if weight:
                weights = self.getColValues('I', applyBeam=applyBeam)
                if weights.shape != vals.shape:
                    weights = np.resize(weights, vals.shape)
                weightCol = Column(name='Weight', data=weights)
                valWeightCol = Column(name='ValWeight', data=vals*weights)
                self.table.add_column(valWeightCol)
                self.table.add_column(weightCol)
                numer = self.table['ValWeight'].groups.aggregate(npsum).data
                denom = self.table['Weight'].groups.aggregate(npsum).data
                self.table.remove_column('ValWeight')
                self.table.remove_column('Weight')
            else:
                valCol = Column(name='Val', data=vals)
                self.table.add_column(valCol)
                numer = self.table['Val'].groups.aggregate(npsum).data
                self.table.remove_column('Val')

            return Column(name=colName, data=np.array(numer/denom),
                          unit=self.table[colName].unit)
        else:
            def npavg(c):
                return np.average(c, axis=0)

            return self.table[colName].groups.aggregate(npavg)

    def _getSizeColumn(self, weight=True, applyBeam=False):
        """
        Returns column of source largest angular sizes.

        Parameters
        ----------
        weight : bool, optional
            If True, return size weighted by flux
        applyBeam : bool, optional
            If True, fluxes will be attenuated by the beam

        Returns
        -------
        col : astropy Column
            Column object with sizes from MajorAxis or from aggregated values if
            the model has patches

        """
        from astropy.table import Column
        import numpy as np

        if weight:
            method = 'wmean'
        else:
            method = 'mean'

        if self.hasPatches:
            # Get patch positions
            RAAvg, DecAvg = self.getPatchPositions(method=method, asArray=True,
                                                   applyBeam=applyBeam)

            # Fill out the columns by repeating the average value over the
            # entire group
            RAAvgFull = np.zeros(len(self.table), dtype=float)
            DecAvgFull = np.zeros(len(self.table), dtype=float)
            for i, ind in enumerate(self.table.groups.indices[1:]):
                RAAvgFull[self.table.groups.indices[i]: ind] = RAAvg[i]
                DecAvgFull[self.table.groups.indices[i]: ind] = DecAvg[i]

            dist = self._calculateSeparation(self.table['Ra'],
                                             self.table['Dec'], RAAvgFull, DecAvgFull)
            if weight:
                if applyBeam and self._hasBeam:
                    appFluxes = self.getColValues('I', applyBeam=True)
                    weightCol = Column(name='Weight', data=appFluxes)
                    valWeightCol = Column(name='ValWeight', data=dist*appFluxes)
                else:
                    weightCol = Column(name='Weight', data=self.table['I'].data)
                    valWeightCol = Column(name='ValWeight', data=dist*self.table['I'].data)
                self.table.add_column(valWeightCol)
                self.table.add_column(weightCol)
                numer = self.table['ValWeight'].groups.aggregate(np.sum).data * 2.0
                denom = self.table['Weight'].groups.aggregate(np.sum).data
                self.table.remove_column('ValWeight')
                self.table.remove_column('Weight')
                col = Column(name='Size', data=numer/denom,
                             unit='degree')
            else:
                valCol = Column(name='Val', data=dist)
                self.table.add_column(valCol)
                size = self.table['Val'].groups.aggregate(np.max).data * 2.0
                self.table.remove_column('Val')
                col = Column(name='Size', data=size, unit='degree')
        else:
            if 'majoraxis' in self.table.colnames:
                col = self.table['MajorAxis']
            else:
                col = Column(name='Size', data=np.zeros(len(self.table)), unit='degree')

        if hasattr(col, 'filled'):
            outcol = col.filled(fill_value=0.0)
        else:
            outcol = col
        outcol.convert_unit_to('arcsec')

        return outcol

    def _calculateSeparation(self, ra1, dec1, ra2, dec2):
        """
        Returns angular separation between two coordinates (all in degrees)

        Parameters
        ----------
        ra1 : float or array
            RA of coordinate 1 in degrees
        dec1 : float or array
            Dec of coordinate 1 in degrees
        ra2 : float
            RA of coordinate 2 in degrees
        dec2 : float
            Dec of coordinate 2 in degrees

        Returns
        -------
        separation : Angle object or array
            Angular separation in degrees

        """
        from .operations_lib import calculateSeparation

        return calculateSeparation(ra1, dec1, ra2, dec2)

    def getDistance(self, RA, Dec, byPatch=False, units=None):
        """
        Returns angular distance for each source or patch to specified position

        Parameters
        ----------
        RA : float or str
            RA of position to which the distance is desired (in degrees or
            makesourcedb format)
        Dec : float or str
            Dec of position to which the distance is desired (in degrees or
            makesourcedb format)
        byPatch : bool, optional
            Calculate distance by patches instead of by sources
        units : str, optional
            Units for resulting distance. If None, units are degrees

        Returns
        -------
        dist : array
            Array of distances

        Examples
        --------
        Find distance in degrees to a position for all sources::

            >>> s.getDistance(94.0, 42.0)

        Find distance in arcmin::

            >>> s.getDistance(94.0, 42.0, units='arcmin')

        Find distance to patch centers:

            >>> s.setPatchPositions(method='mid')
            >>> s.getDistance(94.0, 42.0, byPatch=True)

        """
        from .tableio import RA2Angle, Dec2Angle

        if byPatch and self.hasPatches:
            # Get patch positions
            sRA, sDec = self.getPatchPositions(asArray=True)
        else:
            sRA = self.getColValues('RA')
            sDec = self.getColValues('Dec')

        RA = RA2Angle(RA)
        Dec = Dec2Angle(Dec)

        dist = self._calculateSeparation(sRA, sDec, RA, Dec)
        if units is not None:
            return dist.to(units).value
        else:
            return dist.value

    def write(self, fileName=None, format='makesourcedb', clobber=False,
              sortBy=None, lowToHigh=False, addHistory=True, applyBeam=False,
              invertBeam=False, adjustSI=False):
        """
        Writes the sky model to a file.

        Parameters
        ----------
        filename : str
            Name of output file.
        format: str, optional
            Format of the output file. Allowed formats are:
                - 'makesourcedb' (BBS format)
                - 'fits'
                - 'votable'
                - 'hdf5'
                - 'ds9'
                - 'kvis'
                - 'casa'
                - 'factor'
                - plus all other formats supported by the astropy.table package
        clobber : bool, optional
            If True, an existing file is overwritten.
        sortBy : str or list of str, optional
            Name of columns to sort on. If None, no sorting is done. If
            a list is given, sorting is done on the columns in the order given.
        lowToHigh : bool, optional
            If True, sort values from low to high instead of high to low.
        addHistory : bool, optional
            If True, the history of operations is written to the sky model
            header.
        applyBeam : bool, optional
            If True, fluxes will be adjusted for the beam before being written.
        invertBeam : bool, optional
            If True, the beam correction is inverted (i.e., from apparent sky to
            true sky).
        adjustSI : bool, optional
            If True, adjust the spectral index column for the beam (only works if
            the spectral index in non-logarithmic. I.e., the 'LogarithmicSI' column
            entries are all False)

        Examples
        --------
        Write the model to a makesourcedb sky model file suitable for use with
        BBS::

            >>> s.write('modsky.model')

        Write to a fits catalog::

            >>> s.write('sky.fits', format='fits')

        Write to a ds9 region file::

            >>> s.write('sky.reg', format='ds9')

        """
        import os
        import numpy as np
        from .operations_lib import attenuate

        if fileName is None:
            if self._fileName is None:
                raise IOError("No file name specified.")
            else:
                fileName = self._fileName

        if os.path.exists(fileName):
            if clobber:
                os.remove(fileName)
            else:
                raise IOError("The output file '{0}' exists and clobber = False.".
                              format(fileName))

        table = self.table.copy()

        # Apply beam attenuation
        if applyBeam:
            if 'LogarithmicSI' in self.getColNames():
                if np.any(self.getColValues('LogarithmicSI') == "true"):
                    adjustSI = False
            else:
                # Default is LogarithmicSI=true
                adjustSI = False
            I_orig = self.getColValues('I')
            RADeg = self.getColValues('Ra')
            DecDeg = self.getColValues('Dec')
            if adjustSI:
                spectralIndex = self.getColValues('SpectralIndex')
                referenceFrequency = self.getColValues('ReferenceFrequency')
                I_adj, SI_adj = attenuate(self.beamMS, I_orig, RADeg, DecDeg,
                                          timeIndx=self.beamTime, invert=invertBeam,
                                          spectralIndex=spectralIndex,
                                          referenceFrequency=referenceFrequency)
            else:
                I_adj = attenuate(self.beamMS, I_orig, RADeg, DecDeg,
                                  timeIndx=self.beamTime, invert=invertBeam)
            units = self.table.columns['I'].unit
            table['I'] = I_adj
            table.columns['I'].unit = units
            if adjustSI:
                table['SpectralIndex'] = SI_adj

        # Sort if desired. For 'factor' output, save the order of patches in the
        # table meta
        if sortBy is not None:
            colName = self._verifyColName(sortBy)
            if format.lower() == 'factor' and self.hasPatches:
                indx = np.argsort(self.getColValues('I', aggregate='sum'))
                if not lowToHigh:
                    indx = indx[::-1]
                table.meta['patch_order'] = indx
            else:
                indx = table.argsort(colName)
                if not lowToHigh:
                    indx = indx[::-1]
                table = table[indx]

        if addHistory:
            table.meta['History'] = self.history

        # Add patch sizes in degrees
        if format.lower() == 'factor' and self.hasPatches:
            table.meta['patch_size'] = self.getPatchSizes(units='deg')

        # Add patch fluxes in mJy
        if format.lower() == 'factor' and self.hasPatches:
            table.meta['patch_flux'] = self.getColValues('I', aggregate='sum',
                                                         units='mJy')

        if format.lower() != 'makesourcedb' and format.lower() != 'factor':
            table.meta = {}
        table.write(fileName, format=format.lower())

    def broadcast(self):
        """
        Sends the model to another application using SAMP.

        Both the SAMP hub and the receiving application must be running before
        the table is broadcasted. Examples of SMAP-aware applications are
        TOPCAT, Aladin, and ds9.

        Examples
        --------
        Send the model to TOPCAT. First, start TOPCAT, then run the command::

            >>> s.broadcast()

        TOPCAT should then load the table.

        """
        import tempfile

        tfile = tempfile.NamedTemporaryFile()
        self.table.write(tfile, format='votable')
        tableio.broadcastTable(tfile.name)
        tfile.close()

    def _clean(self):
        """
        Removes duplicate entries.
        """
        names = self.getColValues('Name')
        nameSet = set(names)
        if len(names) == len(nameSet):
            return

        filtNames = []
        filtIndices = []
        for i, name in enumerate(self.getColValues('Name')):
            if name in filtNames:
                filtIndices.append(i)
            else:
                filtNames.append(name)
        nRowsOrig = len(self.table)
        self.table = self.table[filtIndices]
        nRowsNew = len(self.table)
        if nRowsOrig-nRowsNew > 0:
            self.log.info('Removed {0} duplicate sources.'.format(nRowsOrig-nRowsNew))
        self._updateGroups()

    def select(self, filterExpression, aggregate=None, applyBeam=False,
               useRegEx=False, force=True):
        """
        Filters the sky model, keeping all sources that meet the given expression.

        After filtering, the sky model contains only those sources for which the
        given filter expression is true.

        Parameters
        ----------
        filterExpression : str, dict, list, or numpy array
            - If string:
                A string specifying the filter expression in the form:
                '<property> <operator> <value> [<units>]'
                (e.g., 'I <= 10.5 Jy').

            - If dict:
                The filter can also be given as a dictionary in the form:
                {'filterProp':property, 'filterOper':operator,
                    'filterVal':value, 'filterUnits':units}

            - If list:
                The filter can also be given as a list of:
                [property, operator, value] or
                [property, operator, value, units]

            - If numpy array:
                The indices to filter on can be specified directly as a numpy array
                of row or patch indices such as:
                array([ 0,  2, 19, 20, 31, 37])

                or as a numpy array of bools with the same length as the sky model.

                If a numpy array is given and the indices correspond to patches, then
                set aggregate=True.

            The property to filter on must be one of the following:
                - a valid column name
                - the filename of a mask image

            Supported operators are:
                - !=
                - <=
                - >=
                - >
                - <
                - = (or '==')
            Units are optional and must be specified as required by astropy.units.
        aggregate : str, optional
            If set, the selection will be done on values aggregated
            over the patch members. The following aggregation functions are
            available:
                - 'sum': sum of patch values
                - 'mean': mean of patch values
                - 'wmean': Stokes I weighted mean of patch values
                - 'min': minimum of patch values
                - 'max': maximum of patch values
                - True: only valid when the filter indices are specified directly
                as a numpy array. If True, filtering is done on patches instead
                of sources.
        applyBeam : bool, optional
            If True, apparent fluxes will be used.
        useRegEx : bool, optional
            If True, string matching will use regular expression matching. If
            False, string matching uses Unix filename matching.
        force : bool, optional
            If True, selections that result in empty sky models are allowed. If
            False, such selections are not applied and the sky model is unaffected.

        Examples
        --------
        Filter on column 'I' (Stokes I flux). This filter will select all sources
        with Stokes I flux greater than 1.5 Jy::

            >>> s.select('I > 1.5 Jy')
            INFO: Kept 1102 sources.

        If the sky model has patches and the filter is desired per patch, use
        ``aggregate = function``. For example, to select on the sum of the patch
        fluxes::

            >>> s.select('I > 1.5 Jy', aggregate='sum')

        Or, to filter on patches smaller than 5 arcmin in size::

            >>> sizes = s.getPatchSizes(units='arcmin')
            >>> s.select(sizes < 5.0, aggregate=True)

        Filter on source names, keeping those that match "src*_1?"::

            >>> s.select('Name == src*_1?')

        Use a CASA clean mask image named 'clean_mask.mask' to select sources
        that lie in masked regions::

            >>> s.select('clean_mask.mask == True')

        """
        operations.select.select(self, filterExpression, aggregate=aggregate,
                                 applyBeam=applyBeam, useRegEx=useRegEx, force=force)

    def remove(self, filterExpression, aggregate=None, applyBeam=None,
               useRegEx=False, force=True):
        """
        Filters the sky model, removing all sources that meet the given expression.

        After filtering, the sky model contains only those sources for which the
        given filter expression is false.

        Parameters
        ----------
        filterExpression : str, dict, list, or numpy array
            - If string:
                A string specifying the filter expression in the form:
                '<property> <operator> <value> [<units>]'
                (e.g., 'I <= 10.5 Jy').

            - If dict:
                The filter can also be given as a dictionary in the form:
                {'filterProp':property, 'filterOper':operator,
                    'filterVal':value, 'filterUnits':units}

            - If list:
                The filter can also be given as a list of:
                [property, operator, value] or
                [property, operator, value, units]

            - If numpy array:
                The indices to filter on can be specified directly as a numpy array
                of row or patch indices such as:
                array([ 0,  2, 19, 20, 31, 37])

                or as a numpy array of bools with the same length as the sky model.

                If a numpy array is given and the indices correspond to patches, then
                set aggregate=True.

            The property to filter on must be one of the following:
                - a valid column name
                - the filename of a mask image

            Supported operators are:
                - !=
                - <=
                - >=
                - >
                - <
                - = (or '==')
            Units are optional and must be specified as required by astropy.units.
        aggregate : str, optional
            If set, the selection will be done on values aggregated
            over the patch members. The following aggregation functions are
            available:
            - 'sum': sum of patch values
            - 'mean': mean of patch values
            - 'wmean': Stokes I weighted mean of patch values
            - 'min': minimum of patch values
            - 'max': maximum of patch values
            - True: only valid when the filter indices are specified directly as
            a numpy array. If True, filtering is done on patches instead of
            sources.
        applyBeam : bool, optional
            If True, apparent fluxes will be used.
        useRegEx : bool, optional
            If True, string matching will use regular expression matching. If
            False, string matching uses Unix filename matching.
        force : bool, optional
            If True, filters that result in empty sky models are allowed. If
            False, such filters are not applied and the sky model is unaffected.

        Examples
        --------
        Filter on column 'I' (Stokes I flux). This filter will remove all sources
        with Stokes I flux greater than 1.5 Jy::

            >>> s.remove('I > 1.5 Jy')
            INFO: Removed 1102 sources.

        If the sky model has patches and the filter is desired per patch, use
        ``aggregate = function``. For example, to filter on the sum of the patch
        fluxes::

            >>> s.remove('I > 1.5 Jy', aggregate='sum')

        Or, to filter on patches smaller than 5 arcmin in size::

            >>> sizes = s.getPatchSizes(units='arcmin')
            >>> s.remove(sizes < 5.0, aggregate=True)

        Filter on source names, removing those that match "src*_1?" (e.g.,
        'src2345_15', 'srcB2_1a', etc.)::

            >>> s.remove('Name == src*_1?')

        Use a CASA clean mask image named 'clean_mask.mask' to remove sources
        that lie in masked regions::

            >>> s.remove('clean_mask.mask == True')

        """
        operations.remove.remove(self, filterExpression, aggregate=aggregate,
                                 applyBeam=applyBeam, useRegEx=useRegEx, force=force)

    def group(self, algorithm, targetFlux=None, weightBySize=False, numClusters=100, FWHM=None,
              threshold=0.1, applyBeam=False, root='Patch', pad_index=False,
              method='mid', facet="", byPatch=False, kernelSize=0.1,
              nIterations=100, lookDistance=0.2, groupingDistance=0.01):
        """
        Groups sources into patches.

        Parameters
        ----------
        LSM : SkyModel object
            Input sky model.
        algorithm : str
            Algorithm to use for grouping:
            - 'single' => all sources are grouped into a single patch
            - 'every' => every source gets a separate patch named 'source_patch'
            - 'cluster' => SAGECAL clustering algorithm that groups sources into
                specified number of clusters (specified by the numClusters parameter)
            - 'tessellate' => group into tiles whose total flux approximates
                the target flux (specified by the targetFlux parameter)
            - 'threshold' => group by convolving the sky model with a Gaussian beam
                and then thresholding to find islands of emission (NOTE: all sources
                are currently considered to be point sources of flux unity)
            - 'facet' => group by facets using as an input a fits file. It requires
                the use of the additional parameter 'facet' to enter the name of the
                fits file.
            - 'voronoi' => given a previously grouped sky model, voronoi tesselate
                using the patch positions for patches above the target flux (specified
                by the targetFlux parameter)
            - 'meanshift' => use the meanshift clustering algorithm
            - the filename of a mask image => group by masked regions (where mask =
                True). Sources outside of masked regions are given patches of their
                own
        targetFlux : str or float, optional
            Target flux for tessellation (the total flux of each tile will be close
            to this value) and voronoi algorithms. The target flux can be specified
            as either a float in Jy or as a string with units (e.g., '25.0 mJy')
        weightBySize : bool, optional
            If True, fluxes are weighted by patch size (as median_size / size) when
            the targetFlux criterion is applied. Patches with sizes below the median
            (flux-weighted) size are upweighted and those above the mean are
            downweighted
        numClusters : int, optional
            Number of clusters for clustering. Sources are grouped around the
            numClusters brightest sources.
        FWHM : str or float, optional
            FWHM of convolving Gaussian used for thresholding. The FWHM can
            be specified as either a float in degrees or as a string with units
            (e.g., '25.0 arcsec')
        threshold : float, optional
            Value between 0 and 1 above which emission is considered for thresholding
        applyBeam : bool, optional
            If True, fluxes will be attenuated by the beam.
        root : str, optional
            Root string from which patch names are constructed. For 'single', the
            patch name will be set to root; for the other grouping algorithms, the
            patch names will be 'root_INDX', where INDX is an integer ranging from
            (0:nPatches).
        pad_index : bool, optional
            If True, pad the INDX is used in the patch names. E.g., facet_patch_001
            instead of facet_patch_1
        method : None or str, optional
            This parameter specifies the method used to set the patch positions:
            - 'mid' => the position is set to the midpoint of the patch
            - 'mean' => the positions is set to the mean RA and Dec of the patch
            - 'wmean' => the position is set to the flux-weighted mean RA and
            Dec of the patch
            - 'zero' => set all positions to [0.0, 0.0]
        facet : str, optional
            Facet fits file used with the algorithm 'facet'
        byPatch : bool, optional
            For the 'tessellate' or 'meanshift' algorithms, use patches instead of
            sources
        kernelSize : float, optional
            Kernel size in degrees for 'meanshift' grouping
        nIterations : int, optional
            Number of iterations for 'meanshift' grouping
        lookDistance : float, optional
            Look distance in degrees for 'meanshift' grouping
        groupingDistance : float, optional
            Grouping distance in degrees for 'meanshift' grouping

        Examples
        --------
        Tesselate the sky model into patches with approximately 30 Jy total
        flux:

            >>> s.group('tessellate', targetFlux=30.0)

        """
        operations.group.group(self, algorithm, targetFlux=targetFlux, weightBySize=weightBySize,
                               numClusters=numClusters, FWHM=FWHM, threshold=threshold,
                               applyBeam=applyBeam, root=root, pad_index=pad_index,
                               method=method, facet=facet, byPatch=byPatch,
                               kernelSize=kernelSize, nIterations=nIterations,
                               lookDistance=lookDistance,
                               groupingDistance=groupingDistance)

    def transfer(self, patchSkyModel, matchBy='name', radius=0.1):
        """
        Transfer patches from the input sky model.

        Sources matching those in patchSkyModel will be grouped into
        the patches defined in patchSkyModel. Sources that do not appear in
        patchSkyModel will be placed into separate patches (one per source).
        Patch positions are not transferred (as they may no longer be appropriate
        after transfer).

        Parameters
        ----------
        patchSkyModel : str or SkyModel object
            Input sky model from which to transfer patches.
        matchBy : str, optional
            Determines how matching sources are determined:
            - 'name' => matches are identified by name
            - 'position' => matches are identified by radius. Sources within the
                radius specified by the radius parameter are considered matches
        radius : float or str, optional
            Radius in degrees (if float) or 'value unit' (if str; e.g., '30 arcsec')
            for matching when matchBy='position'

        Examples
        --------
        Transfer patches from one sky model to another and set their positions
        (matching sources are identified by name)::

            >>> s.transfer('master_sky.model')
            >>> s.setPatchPositions(method='mid')

        """
        operations.transfer.transfer(self, patchSkyModel, matchBy=matchBy,
                                     radius=radius)

    def move(self, name, position=None, shift=None):
        """
        Move or shift a source or sources.

        If both a position and a shift are specified, a source is moved to the
        new position and then shifted. Note that only a single source can be
        moved to a new position. However, multiple sources can be shifted.

        If an xyshift is specified, a FITS file must also be specified to define
        the WCS system. If a position, a shift, and an xyshift are all specified,
        a source is moved to the new position, shifted in RA and Dec, and then
        shifted in x and y.

        Parameters
        ----------
        LSM : SkyModel object
            Input sky model
        name : str or list
            Source name or list of names (can include wildcards)
        position : list, optional
            A list specifying a new position as [RA, Dec] in either makesourcedb
            format (e.g., ['12:23:43.21', '+22.34.21.2']) or in degrees (e.g.,
            [123.2312, 23.3422])
        shift : list, optional
            A list specifying the shift as [RAShift, DecShift] in degrees (e.g.,
            [0.02312, 0.00342])
        xyshift : list, optional
            A list specifying the shift as [xShift, yShift] in pixels. A FITS file
            must be specified with the fitsFILE argument
        fitsFile : str, optional
            A FITS file from which to take WCS information to transform the pixel
            coordinates to RA and Dec values. The xyshift argument must be specfied
            for this to be useful

        Examples
        --------
        Move source '1609.6+6556' to a new position::

            >>> s.move('1609.6+6556', position=['16:10:00', '+65.57.00'])

        Shift the source by 10 arcsec in Dec::

            >>> s.move('1609.6+6556', shift=[0.0, 10.0/3600.0])

        Shift all sources by 10 pixels in x::

            >>> s.move('*', xyshift=[10, 0], fitsFile='image.fits')

        """
        operations.move.move(self, name, position=position, shift=shift)

    def add(self, colNamesVals):
        """
        Add a source to the sky model.

        Parameters
        ----------
        colNamesVals : dict
            A dictionary that specifies the column values for the source to be
            added

        Examples:
        ---------
        Add a point source::

            >>> source = {'Name':'src1', 'Type':'POINT', 'Ra':'12:32:10.1',
                'Dec':'23.43.21.21', 'I':2.134}
            >>> s.add(source)

        """
        operations.add.add(self, colNamesVals)

    def merge(self, patches, name=None):
        """
        Merge two or more patches together.

        Parameters
        ----------
        patches : list of str
            List of patch names to merge
        name : str, optional
            Name of resulting merged patch. If None, the merged patch uses the
            name of the first patch in the input patches list

        Examples
        --------
        Merge three patches into one named 'binmerged'::

            >>> s.merge(['bin0', 'bin1', 'bin2'], 'binmerged')

         """
        operations.merge.merge(self, patches, name=name)

    def concatenate(self, LSM2, matchBy='name', radius=0.1, keep='all',
                    inheritPatches=False):
        """
        Concatenate two sky models.

        Parameters
        ----------
        LSM2 : str or SkyModel object
            Secondary sky model to concatenate with the parent sky model
        matchBy : str, optional
            Determines how duplicate sources are determined:
            - 'name' => duplicates are identified by name
            - 'position' => duplicates are identified by radius. Sources within the
                radius specified by the radius parameter are considered duplicates
        radius : float or str, optional
            Radius in degrees (if float) or 'value unit' (if str; e.g., '30 arcsec')
            for matching when matchBy='position'
        keep : str, optional
            Determines how duplicates are treated:
            - 'all' => all duplicates are kept; those with identical names are re-
                named
            - 'from1' => duplicates kept are those from sky model 1 (the parent)
            - 'from2' => duplicates kept are those from sky model 2 (the secondary)
        inheritPatches : bool, optional
            If True, duplicates inherit the patch name from the parent sky model. If
            False, duplicates keep their own patch names.

        Examples
        --------
        Concatenate two sky models, identifying duplicates by matching to the source
        names. When duplicates are found, keep the source from the parent sky model
        and discard the duplicate from secondary sky model (this might be useful when
        merging two gsm.py sky models that have some overlap)::

            >>> LSM2 = lsmtool.load('gsm_sky2.model')
            >>> s.concatenate(LSM2, matchBy='name', keep='from1')

        Concatenate two sky models, identifying duplicates by matching to the source
        positions within a radius of 10 arcsec. When duplicates are found, keep the
        source from the secondary sky model and discard the duplicate from the parent
        sky model (this might be useful when replacing parts of a low-resolution
        sky model with a high-resolution one)::

            >>> LSM2 = lsmtool.load('high_res_sky.model')
            >>> s.concatenate(LSM2, matchBy='position', radius=10.0/3600.0,
                keep='from2')

        """
        if type(LSM2) is str:
            LSM2 = SkyModel(LSM2)
        operations.concatenate.concatenate(self, LSM2, matchBy=matchBy,
                                           radius=radius, keep=keep,
                                           inheritPatches=inheritPatches)

    def compare(self, LSM2, radius='10 arcsec', outDir='.', labelBy=None,
                ignoreSpec=None, excludeMultiple=True, excludeByFlux=False, name1=None,
                name2=None, format='pdf', make_plots=True):
        """
        Compare two sky models.

        Comparison plots and a text file with statistics are written out to the
        an output directory. Plots are made for:
            - flux ratio vs. radius from sky model center
            - flux ratio vs. sky position
            - flux ratio vs flux
            - position offsets
        The following statistics are saved to 'stats.txt' in the output directory:
            - mean and standard deviation of flux ratio
            - mean and standard deviation of RA offsets
            - mean and standard deviation of Dec offsets
        These statistics are also returned as a dictionary.

        Parameters
        ----------
        LSM2 : SkyModel object
            Secondary sky model to compare to the parent sky model
        radius : float or str, optional
            Radius in degrees (if float) or 'value unit' (if str; e.g., '30 arcsec')
            for matching
        outDir : str, optional
            Plots are saved to this directory
        labelBy : str, optional
            One of 'source' or 'patch': label points using source names ('source') or
            patch names ('patch')
        ignoreSpec : float, optional
            Ignore sources with this spectral index
        excludeMultiple : bool, optional
            If True, sources with multiple matches are excluded. If False, the
            nearest of the multiple matches will be used for comparison
        excludeByFlux : bool, optional
            If True, matches whose predicted fluxes differ from the parent model
            fluxes by 25% are excluded from the positional offset plot.
        name1 : str, optional
            Name to use in the plots for the primary sky model. If None, 'Model 1' is used.
        name2 : str, optional
            Name to use in the plots for LSM2. If None, 'Model 2' is used.
        format : str, optional
            Format of plot files.
        make_plots : bool, optional
            If True, the plots described above are made.

        Returns
        -------
        stats : dict
            Dict of statistics with the following keys (where the clipped values
            are after 3-sigma clipping):
                - 'meanRatio'
                - 'stdRatio'
                - 'meanRAOffsetDeg'
                - 'stdRAOffsetDeg'
                - 'meanDecOffsetDeg'
                - 'stdDecOffsetDeg'
                - 'meanClippedRatio'
                - 'stdClippedRatio'
                - 'meanClippedRAOffsetDeg'
                - 'stdClippedRAOffsetDeg'
                - 'meanClippedDecOffsetDeg'
                - 'stdClippedDecOffsetDeg'

        Examples
        --------
        Compare two sky models and save plots::

            >>> LSM2 = lsmtool.load('sky2.model')
            >>> s.compare(LSM2, outDir='comparison_results/')

        Compare a LOFAR sky model to a global sky model made from VLSS+TGSS+NVSS (where
        refRA and refDec are the approximate center of the LOFAR sky model coverage)::

            >>> LSM2 = lsmtool.load('GSM', VOPosition=[refRA, refDec], VORadius='5 deg')
            >>> s.compare(LSM2, radius='30 arcsec', excludeMultiple=True,
                outDir='comparison_results/', name1='LOFAR', name2='GSM', format='png')

        """
        if type(LSM2) is str:
            LSM2 = SkyModel(LSM2)
        stats = operations.compare.compare(self, LSM2, radius=radius, outDir=outDir,
                                           labelBy=labelBy, ignoreSpec=ignoreSpec,
                                           excludeMultiple=excludeMultiple,
                                           excludeByFlux=excludeByFlux, name1=name1,
                                           name2=name2, format=format, make_plots=make_plots)
        return stats

    def plot(self, fileName=None, labelBy=None):
        """
        Shows a simple plot of the sky model.

        The circles in the plot are scaled with flux. If the sky model is grouped
        into patches, sources are colored by patch and the patch positions are
        indicated with stars.

        Parameters
        ----------
        fileName : str, optional
            If given, the plot is saved to a file instead of displayed.
        labelBy : str, optional
            One of 'source' or 'patch': label points using source names ('source')
            or patch names ('patch')

        Examples:
        ---------
        Plot and display to the screen::

            >>> s.plot()

        Plot and save to a PDF file::

            >>> s.plot('sky_plot.pdf')

        """
        operations.plot.plot(self, fileName=fileName, labelBy=labelBy)

    def rasterize(self, cellsize, fileRoot=None, writeRegionFile=False, clobber=False):
        """
        Rasterize the sky model to FITS images (one image per spectral term).

        The resulting images can be used with DDECal in DPPP for prediction using IDG
        (if the sky model is grouped into contiguous patches, a ds9 region file defining
        the Voronio patches can also written).

        Note: currently, only sky models with LogarithmicSI = False are supported at
        this time.

        Parameters
        ----------
        cellsize : float
            The cellsize in degrees for the output image.
        fileRoot : str, optional
            Filename root for the output FITS images. The images will be named
            fileRoot_0.fits, fileRoot_1.fits, etc. (one for each spectral term in
            the sky model). If writeRegionFile is True, a ds9 region file is also
            written as fileRoot.reg.
        writeRegionFile : bool, optional
            If True and the sky model is grouped into contiguous patches, a ds9 region
            file defining the Voronio patches will be written (this file is required
            for DDECal)
        clobber : bool, optional
            If True, existing files are overwritten.
        """
        import numpy as np
        from astropy.io import fits as pyfits
        from astropy import wcs
        from .operations_lib import make_template_image, gaussian_fcn, tessellate, xy2radec, radec2xy

        # Make a blank image for each spectral term
        referenceFrequency = self.getColValues('ReferenceFrequency')
        refFreq = referenceFrequency[0]  # TODO: allow per-source ref freq
        fluxes = self.getColValues('I')
        types = self.getColValues('Type')
        nsources = len(fluxes)
        if 'SpectralIndex' in self.getColNames():
            spectral_indices = self.getColValues('SpectralIndex')
        else:
            spectral_indices = [[]] * nsources
        nterms = len(spectral_indices[0]) + 1

        # Check that LogarithmicSI = False for all entries
        if nterms > 1:
            logsi = self.getColValues('LogarithmicSI')
            if np.any(logsi == 'true'):
                raise RuntimeError('Sky model has one or more sources with '
                                   'LogarithmicSI = True. Only sky models with '
                                   'LogarithmicSI = False are supported at this time.')

        image_names = ['{0}_{1}.fits'.format(fileRoot, i) for i in range(nterms)]
        x, y, refRA, refDec = self._getXY(crdelt=cellsize)
        if 'GAUSSIAN' in types:
            fwhm = np.max(self.getColValues('MajorAxis', units='degree') * cellsize)
            max_source_size = int(np.ceil(fwhm * 1.5))
        else:
            max_source_size = 2
        xpadding = int(0.2 * (np.max(x) - np.min(x)))
        ypadding = int(0.2 * (np.max(y) - np.min(y)))
        xpadding += max_source_size
        if xpadding % 2:
            xpadding += 1
        ypadding += max_source_size
        if ypadding % 2:
            ypadding += 1
        xsize = int(np.max(x) - np.min(x)) + xpadding
        ysize = int(np.max(y) - np.min(y)) + ypadding

        # Now we have the size, refine the RA, Dec of the image center
        xcen = np.min(x) + (np.max(x) - np.min(x)) / 2.0
        ycen = np.min(y) + (np.max(y) - np.min(y)) / 2.0
        refRA, refDec = xy2radec([xcen], [ycen], refRA=refRA, refDec=refDec, crdelt=cellsize)
        RA = self.getColValues('Ra')
        Dec = self.getColValues('Dec')

        for image_name in image_names:
            make_template_image(image_name, refRA[0], refDec[0], refFreq,
                                ximsize=xsize, yimsize=ysize, cellsize_deg=cellsize)

        # Build each image, one at a time (to minimize memory usage)
        for t, image_name in enumerate(image_names):
            # Read in image data
            hdu = pyfits.open(image_name, memmap=False)
            imdata = hdu[0].data
            w = wcs.WCS(hdu[0].header)

            # Loop over sources, adding them to the images (note that the spectral terms
            # sum together when summing polynomials, just as the flux densities do)
            if t == 0:
                # Flux densities
                itervalues = fluxes
            else:
                # Spectral terms
                itervalues = spectral_indices
            for i, (ra_src, dec_src, val, type) in enumerate(zip(RA, Dec, itervalues, types)):
                if t > 0:
                    v = val[t-1]
                    const = True
                else:
                    v = val
                    const = False
                ra_dec = np.array([[ra_src, dec_src, 0.0, 0.0]])
                xs, ys = w.wcs_world2pix(ra_dec, 0)[0][0], w.wcs_world2pix(ra_dec, 0)[0][1]
                if type == 'POINT':
                    imdata[0, 0, int(np.round(ys)), int(np.round(xs))] += v
                elif type == 'GAUSSIAN':
                    S1 = self.getColValues('MajorAxis', units='degree')[i] / cellsize  # pixels
                    S2 = self.getColValues('MinorAxis', units='degree')[i] / cellsize  # pixels
                    Th = self.getColValues('Orientation')[i]  # degrees
                    C1, C2 = ys, xs
                    b = np.ceil(S1 * 2.5)
                    bbox = np.s_[max(0, int(C1-b)):min(xsize, int(C1+b+1)),
                                 max(0, int(C2-b)):min(ysize, int(C2+b+1))]
                    x_ax, y_ax = np.mgrid[bbox]
                    g = [v, C1, C2, S1, S2, Th]
                    ffimg = gaussian_fcn(g, x_ax, y_ax, const=const)
                    imdata[0, 0, :, :][bbox] += ffimg
            hdu[0].data = imdata
            hdu.writeto(image_name, overwrite=True)

        # Make region file if needed
        if writeRegionFile and self.hasPatches:
            RA, Dec = self.getPatchPositions(asArray=True)
            patch_names = self.getPatchNames()
            x_pix_list = []
            y_pix_list = []
            for ra_src, dec_src in zip(RA, Dec):
                ra_dec = np.array([[ra_src, dec_src, 0.0, 0.0]])
                y_pix, x_pix = w.wcs_world2pix(ra_dec, 0)[0][0], w.wcs_world2pix(ra_dec, 0)[0][1]
                x_pix_list.append(x_pix)
                y_pix_list.append(y_pix)
            dist_pix = np.sqrt(xsize**2 + ysize**2)
            vertices = tessellate(x_pix_list, y_pix_list, w, dist_pix)
            lines = []
            lines.append('# Region file format: DS9 version 4.0\nglobal color=green '
                         'font="helvetica 10 normal" select=1 highlite=1 edit=1 '
                         'move=1 delete=1 include=1 fixed=0 source=1\nfk5\n')
            for verts, pname in zip(vertices, patch_names):
                xylist = []
                varray = np.array(verts).T
                RAs = varray[0]
                Decs = varray[1]
                for x, y in zip(RAs, Decs):
                    xylist.append('{0}, {1}'.format(x, y))
                lines.append('polygon({0}) # text={{{1}}}\n'.format(', '.join(xylist), pname))

            outputfile = '{0}.reg'.format(fileRoot)
            with open(outputfile, 'w') as f:
                f.writelines(lines)





# -*- coding: utf-8 -*-
#
# Defines astropy.table reader and writer functions for the following formats
#   - makesourcedb/BBS (reader and writer)
#   - ds9 (writer only)
#   - kvis (writer only)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
import astropy
from astropy.table import Table, Column, MaskedColumn
from astropy.coordinates import Angle
from astropy.io import registry
import astropy.io.ascii as ascii
from distutils.version import LooseVersion
import numpy as np
import numpy.ma as ma
import re
import logging
import os
from copy import deepcopy
from .operations_lib import normalize_ra, normalize_dec

# Python 3 compatibility
try:
    dict.iteritems
except AttributeError:
    # Python 3
    def itervalues(d):
        return iter(d.values())

    def iteritems(d):
        return iter(d.items())
    numpy_type = "U"
else:
    # Python 2
    def itervalues(d):
        return d.itervalues()

    def iteritems(d):
        return d.iteritems()
    numpy_type = "S"
try:
    unicode = unicode
except NameError:
    # Python 3
    basestring = (str, bytes)
else:
    # Python 2
    basestring = basestring
import io
try:
    # Python 2
    file_types = (file, io.IOBase)
except NameError:
    # Python 3
    file_types = (io.IOBase,)

# Define the valid columns here as dictionaries. The entry key is the lower-case
# name of the column, the entry value is the key used in the astropy table of the
# SkyModel object. For details, see:
# https://www.astron.nl/lofarwiki/doku.php?id=public:user_software:documentation:makesourcedb
allowedColumnNames = {'name': 'Name', 'type': 'Type', 'patch': 'Patch',
                      'ra': 'Ra', 'dec': 'Dec', 'i': 'I', 'q': 'Q', 'u': 'U', 'v': 'V',
                      'majoraxis': 'MajorAxis', 'minoraxis': 'MinorAxis',
                      'orientation': 'Orientation', 'orientationisabsolute': 'OrientationIsAbsolute',
                      'ishapelet': 'IShapelet', 'qshapelet': 'QShapelet', 'ushapelet': 'UShapelet',
                      'vshapelet': 'VShapelet', 'category': 'Category', 'logarithmicsi': 'LogarithmicSI',
                      'rotationmeasure': 'RotationMeasure', 'polarizationangle': 'PolarizationAngle',
                      'polarizedfraction': 'PolarizedFraction', 'referencewavelength': 'ReferenceWavelength',
                      'referencefrequency': 'ReferenceFrequency', 'spectralindex': 'SpectralIndex'}

allowedColumnUnits = {'name': None, 'type': None, 'patch': None, 'ra': 'degree',
                      'dec': 'degree', 'i': 'Jy', 'i-apparent': 'Jy', 'q': 'Jy', 'u': 'Jy', 'v': 'Jy',
                      'majoraxis': 'arcsec', 'minoraxis': 'arcsec', 'orientation': 'degree',
                      'orientationisabsolute': None,
                      'ishapelet': None, 'qshapelet': None, 'ushapelet': None,
                      'vshapelet': None, 'category': None, 'logarithmicsi': None,
                      'rotationmeasure': 'rad/m^2', 'polarizationangle': 'rad',
                      'polarizedfraction': 'PolarizedFraction',
                      'referencewavelength': 'ReferenceWavelength',
                      'referencefrequency': 'Hz', 'spectralindex': None}

allowedColumnDefaults = {'name': 'N/A', 'type': 'N/A', 'patch': 'N/A', 'ra': 0.0,
                         'dec': 0.0, 'i': 0.0, 'q': 0.0, 'u': 0.0, 'v': 0.0, 'majoraxis': 0.0,
                         'minoraxis': 0.0, 'orientation': 0.0, 'orientationisabsolute': 'false',
                         'ishapelet': 'N/A', 'qshapelet': 'N/A', 'ushapelet': 'N/A',
                         'vshapelet': 'N/A', 'category': 2, 'logarithmicsi': 'true',
                         'rotationmeasure': 0.0, 'polarizationangle': 0.0,
                         'polarizedfraction': 0.0, 'referencewavelength': 'N/A',
                         'referencefrequency': 0.0, 'spectralindex': [0.0]}

requiredColumnNames = ['Name', 'Type', 'Ra', 'Dec', 'I']

allowedVOServices = {
    'nvss': 'http://vizier.u-strasbg.fr/viz-bin/votable/-A?-source=VIII/65&amp;',
    'wenss': 'http://vizier.u-strasbg.fr/viz-bin/votable/-A?-source=VIII/62A&amp;'}

# Define the various URLs used for downloading sky models
TGSS_URL = 'http://tgssadr.strw.leidenuniv.nl/cgi-bin/gsmv4.cgi'
GSM_URL = 'https://lcs165.lofar.eu/cgi-bin/gsmv1.cgi'
LOTSS_URL = 'https://vo.astron.nl/lotss_dr2/q/src_cone/form'


def raformat(val):
    """
    Column formatter for RA values.

    Parameters
    ----------
    val : float
        Input RA value in deg

    Returns
    -------
    valstr : str
        Formatted string as 'hh:mm:ss.s'

    """
    val = normalize_ra(val)
    return Angle(val, unit='degree').to_string(unit='hourangle', sep=':')


def decformat(val):
    """
    Column formatter for Dec values.

    Parameters
    ----------
    val : float
        Input Dec value in deg

    Returns
    -------
    valstr : str
        Formatted string as 'dd.mm.ss.s'

    """
    val = normalize_dec(val)
    return Angle(val, unit='degree').to_string(unit='degree', sep='.')


def fluxformat(val):
    """
    Column formatter for flux density values.

    Parameters
    ----------
    val : float
        Input flux density value in Jy

    Returns
    -------
    valstr : str
        Formatted string to 3 digits

    """
    if type(val) is ma.core.MaskedConstant:
        return '{}'.format(val)
    else:
        return '{0:0.3f}'.format(val)


def skyModelReader(fileName):
    """
    Reads a makesourcedb sky model file into an astropy table.

    See https://www.astron.nl/lofarwiki/doku.php?id=public:user_software:documentation:makesourcedb
    for details. Note that source names, types, and patch names are limited to
    a length of 100 characters.

    Parameters
    ----------
    fileName : str
        Input ASCII file from which the sky model is read. Must
        respect the makesourcedb format

    Returns
    -------
    table : astropy.table.Table object

    """
    log = logging.getLogger('LSMTool.Load')

    # Open the input file
    try:
        modelFile = open(fileName)
        log.debug('Reading {0}'.format(fileName))
    except IOError as e:
        raise IOError('Could not open {0}: {1}'.format(fileName, e.strerror))

    # Read format line
    formatString = None
    for line in modelFile:
        if 'format' in line.lower():
            formatString = line
            break
    modelFile.close()
    if formatString is None:
        raise IOError("No valid format line found in file '{0}'.".format(fileName))

    # Process the header
    colNames, hasPatches, colDefaults, metaDict = processFormatString(formatString)

    # Read model into astropy table object
    outlines = []
    log.debug('Reading file...')
    with open(fileName) as f:
        for line in f:
            outline, metaDict = processLine(line, metaDict, colNames)
            if outline is not None:
                outlines.append(outline)
    outlines.append('\n')  # needed in case of single-line sky models

    # Create table
    table = createTable(outlines, metaDict, colNames, colDefaults)

    return table


def createTable(outlines, metaDict, colNames, colDefaults):
    """
    Creates an astropy table from inputs.

    Parameters
    ----------
    outlines : list of str
        Input lines
    metaDict : dict
        Input meta data
    colNames : list of str
        Input column names
    colDefaults : list
        Input column default values

    Returns
    -------
    table : astropy.table.Table object

    """
    log = logging.getLogger('LSMTool.Load')

    # Before loading table into an astropy Table object, set lengths of Name,
    # Patch, and Type columns to 100 characters to ensure long names are not
    # truncated. The LogarithmicSI and OrientationIsAbsolute columns are set
    # to 5 characters to allow true/false values to be stored as strings without
    # truncation. Due to a change in the astropy table API with v4.1, we have to
    # check the version and use the appropriate column names
    if LooseVersion(astropy.__version__) < LooseVersion('4.1'):
        # Use the input column names for the converters
        nameCol = 'col{0}'.format(colNames.index('Name')+1)
        typeCol = 'col{0}'.format(colNames.index('Type')+1)
        if 'Patch' in colNames:
            patchCol = 'col{0}'.format(colNames.index('Patch')+1)
        if 'LogarithmicSI' in colNames:
            logSICol = 'col{0}'.format(colNames.index('LogarithmicSI')+1)
        if 'OrientationIsAbsolute' in colNames:
            orienCol = 'col{0}'.format(colNames.index('OrientationIsAbsolute')+1)
    else:
        # Use the output column names for the converters
        nameCol = 'Name'
        typeCol = 'Type'
        patchCol = 'Patch'
        logSICol = 'LogarithmicSI'
        orienCol = 'OrientationIsAbsolute'
    converters = {}
    converters[nameCol] = [ascii.convert_numpy('{}100'.format(numpy_type))]
    converters[typeCol] = [ascii.convert_numpy('{}100'.format(numpy_type))]
    if 'Patch' in colNames:
        converters[patchCol] = [ascii.convert_numpy('{}100'.format(numpy_type))]
    if 'LogarithmicSI' in colNames:
        converters[logSICol] = [ascii.convert_numpy('{}5'.format(numpy_type))]
    if 'OrientationIsAbsolute' in colNames:
        converters[orienCol] = [ascii.convert_numpy('{}5'.format(numpy_type))]

    log.debug('Creating table...')
    table = Table.read('\n'.join(outlines), guess=False, format='ascii.no_header', delimiter=',',
                       names=colNames, comment='#', data_start=0, converters=converters)

    # Convert spectral index values from strings to arrays.
    if 'SpectralIndex' in table.keys():
        log.debug('Converting spectral indices...')
        specOld = table['SpectralIndex'].data.tolist()
        specVec = []
        maskVec = []
        maxLen = 0
        for entry in specOld:
            try:
                if type(entry) is float or type(entry) is int:
                    maxLen = 1
                else:
                    specEntry = [float(f) for f in entry.split(';')]
                    if len(specEntry) > maxLen:
                        maxLen = len(specEntry)
            except:
                pass
        defSpeclen = len(colDefaults[colNames.index('SpectralIndex')])
        if defSpeclen > maxLen:
            maxLen = defSpeclen
        log.debug('Maximum number of spectral-index terms in model: {0}'.format(maxLen))
        for entry in specOld:
            try:
                # Take existing entry and fix type
                if type(entry) is float or type(entry) is int:
                    specEntry = [float(entry)]
                    specMask = [False]
                else:
                    specEntry = [float(f) for f in entry.split(';')]
                    specMask = [False] * len(specEntry)
            except:
                # No entry in table, so use default value
                specEntry = colDefaults[colNames.index('SpectralIndex')]
                specMask = [False] * len(specEntry)
            while len(specEntry) < maxLen:
                # Add masked values to any entries that are too short
                specEntry.append(0.0)
                specMask.append(True)
            specVec.append(specEntry)
            maskVec.append(specMask)
        specCol = MaskedColumn(name='SpectralIndex', data=np.array(specVec, dtype=float))
        specCol.mask = maskVec
        specIndx = table.keys().index('SpectralIndex')
        table.remove_column('SpectralIndex')
        table.add_column(specCol, index=specIndx)

    # Convert RA and Dec to Angle objects
    log.debug('Converting RA...')
    RARaw = table['Ra'].data.tolist()
    RACol = Column(name='Ra', data=RA2Angle(RARaw))
    RACol.format = raformat
    RAIndx = table.keys().index('Ra')
    table.remove_column('Ra')
    table.add_column(RACol, index=RAIndx)

    log.debug('Converting Dec...')
    DecRaw = table['Dec'].data.tolist()
    DecCol = Column(name='Dec', data=Dec2Angle(DecRaw))
    DecCol.format = decformat
    DecIndx = table.keys().index('Dec')
    table.remove_column('Dec')
    table.add_column(DecCol, index=DecIndx)

    table.columns['I'].format = fluxformat

    # Set column units and default values
    for i, colName in enumerate(colNames):
        log.debug("Setting units for column '{0}' to {1}".format(
            colName, allowedColumnUnits[colName.lower()]))
        table.columns[colName].unit = allowedColumnUnits[colName.lower()]

        if hasattr(table.columns[colName], 'filled') and colDefaults[i] is not None:
            fillVal = colDefaults[i]
            log.debug("Setting default value for column '{0}' to {1}".
                      format(colName, fillVal))
            if colName == 'SpectralIndex':
                # We cannot set the fill value to a list/array, so just use a float
                fillVal = 0.0
            table.columns[colName].set_fill_value(fillVal)
    table.meta = metaDict

    return table


def processFormatString(formatString):
    """
    Proccesses the header string.

    Parameters
    ----------
    formatString : str
        Header line

    Returns
    -------
    colNames : list of str
        Output column names
    hasPatches : bool
        Flag for patches
    colDefaults : dict
        Default values
    metaDict : dict
        Output meta data

    """
    formatString = formatString.strip()
    formatString = formatString.strip('# ')
    if formatString.lower().endswith('format'):
        parts = formatString.split('=')[:-1]
        formatString = 'FORMAT = ' + '='.join(parts).strip('# ()')
    elif formatString.lower().startswith('format'):
        parts = formatString.split('=')[1:]
        formatString = 'FORMAT = ' + '='.join(parts).strip('# ()')
    else:
        raise IOError("Format line not understood.")

    # Check whether sky model has patches
    if 'Patch' in formatString:
        hasPatches = True
    else:
        hasPatches = False

    # Get column names and default values. Non-string columns have default
    # values of 0.0 unless a different value is given in the header.
    if ',' not in formatString:
        raise IOError("Sky model must use ',' as a field separator.")
    colNames = formatString.split(',')

    # Check if a default value in the format string is a list. If it is, make
    # sure the list is complete
    cnStart = None
    cnEnd = None
    for cn in colNames:
        if '[' in cn and ']' not in cn:
            cnStart = cn
        if ']' in cn and '[' not in cn:
            cnEnd = cn
    if cnStart is not None:
        indx1 = colNames.index(cnStart)
        indx2 = colNames.index(cnEnd)
        colNamesFixed = []
        toJoin = []
        for i, cn in enumerate(colNames):
            if i < indx1:
                colNamesFixed.append(cn)
            elif i >= indx1 and i <= indx2:
                toJoin.append(cn)
                if i == len(colNames)-1:
                    colNamesFixed.append(','.join(toJoin))
            elif i > indx2:
                if i == indx2 + 1:
                    colNamesFixed.append(','.join(toJoin))
                    colNamesFixed.append(cn)
                else:
                    colNamesFixed.append(cn)
        colNames = colNamesFixed

    # Now get the defaults
    colDefaults = [None] * len(colNames)
    metaDict = {}
    colNames[0] = colNames[0].split('=')[1]
    for i in range(len(colNames)):
        parts = colNames[i].split('=')
        colName = parts[0].strip().lower()
        if len(parts) == 2:
            try:
                if '[' in parts[1]:
                    # Default is a list
                    defParts = parts[1].strip("'[]").split(',')
                    defaultVal = []
                    for p in defParts:
                        defaultVal.append(float(p.strip()))
                elif 'true' in parts[1].lower():
                    defaultVal = True
                elif 'false' in parts[1].lower():
                    defaultVal = False
                else:
                    defaultVal = float(parts[1].strip("'"))
            except ValueError:
                defaultVal = None
        else:
            defaultVal = None

        if colName == '':
            raise IOError('Skipping of columns is not yet supported.')
        if colName not in allowedColumnNames:
            raise IOError("Column '{0}' is not currently allowed".format(colName))
        else:
            colNames[i] = allowedColumnNames[colName]
            if defaultVal is not None:
                colDefaults[i] = defaultVal
                metaDict[colNames[i]] = defaultVal
            elif allowedColumnDefaults[colName] is not None:
                # Note: we used deepcopy() here to ensure that the original
                # is not altered by later changes to colDefaults
                colDefaults[i] = deepcopy(allowedColumnDefaults)[colName]

    # Check for required columns
    for reqCol in requiredColumnNames:
        if reqCol not in colNames:
            raise IOError("Sky model must have a '{0}' column.".format(reqCol))

    return colNames, hasPatches, colDefaults, metaDict


def processLine(line, metaDict, colNames):
    """
    Processes a makesourcedb line.

    Parameters
    ----------
    line : str
        Data line
    metaDict : dict
        Input meta data
    colNames : list of str
        Input column names

    Returns
    -------
    line : str
        Processed line
    metaDict : dict
        Output meta data

    """
    if line.lower().startswith("format") or line.startswith("#"):
        return None, metaDict

    # Check for SpectralIndex entries, which are unreadable as they use
    # the same separator for multiple orders as used for the columns
    line = line.strip('\n')
    a = re.search(r'\[.*\]', line)
    if a is not None:
        b = line[a.start(): a.end()]
        c = b.strip('[]')
        if ',' in c:
            c = c.replace(',', ';')
        line = line.replace(b, c)
    colLines = line.split(',')

    # Check for patch lines as any line with an empty Name entry. If found,
    # store patch positions in the table meta data.
    nameIndx = colNames.index('Name')
    if colLines[nameIndx].strip() == '':
        if len(colLines) > 4:
            patchIndx = colNames.index('Patch')
            patchName = colLines[patchIndx].strip()
            RAIndx = colNames.index('Ra')
            if colLines[RAIndx].strip() == '':
                patchRA = [0.0]
            else:
                patchRA = RA2Angle(colLines[RAIndx].strip())
            DecIndx = colNames.index('Dec')
            if colLines[DecIndx].strip() == '':
                patchDec = [0.0]
            else:
                patchDec = Dec2Angle(colLines[DecIndx].strip())
            metaDict[patchName] = [patchRA[0], patchDec[0]]
        return None, metaDict

    while len(colLines) < len(colNames):
        colLines.append(' ')

    return ','.join(colLines), metaDict


def RA2Angle(RA):
    """
    Returns Angle objects for input RA values.

    Parameters
    ----------
    RA : str, float or list of str, float
        Values of RA to convert. Can be strings in makesourcedb format or floats
        in degrees.

    Returns
    -------
    RAAngle : astropy.coordinates.Angle object

    """
    import astropy.units as u

    if type(RA) is not list:
        RA = [RA]

    if type(RA[0]) is str:
        try:
            RAAngle = Angle(Angle(RA, unit=u.hourangle), unit=u.deg)
        except KeyboardInterrupt:
            raise
        except Exception as e:
            raise ValueError('RA not understood (must be string in '
                             'makesourcedb format or float in degrees): {0}'.format(e))
    else:
        RA = [normalize_ra(r) for r in RA]
        RAAngle = Angle(RA, unit=u.deg)
    RAAngle.wrap_at('360d', inplace=True)  # wrap to [0 - 360)

    return RAAngle


def Dec2Angle(Dec):
    """
    Returns Angle objects for input Dec values.

    Parameters
    ----------
    Dec : str, float or list of str, float
        Values of Dec to convert. Can be strings in makesourcedb format or floats
        in degrees

    Returns
    -------
    DecAngle : astropy.coordinates.Angle object

    """
    import astropy.units as u

    if type(Dec) is not list:
        Dec = [Dec]

    if type(Dec[0]) is str:
        try:
            DecAngle = Angle(Dec, unit=u.deg)
        except KeyboardInterrupt:
            raise
        except ValueError:
            try:
                DecSex = [decstr.replace('.', ':', 2) for decstr in Dec]
                DecAngle = Angle(DecSex, unit=u.deg)
            except Exception as e:
                raise ValueError('Dec not understood (must be string in '
                                 'makesourcedb format or float in degrees): {0}'.format(e))
        except Exception as e:
            raise ValueError('Dec not understood (must be string in '
                             'makesourcedb format or float in degrees): {0}'.format(e))
    else:
        Dec = [normalize_dec(d) for d in Dec]
        DecAngle = Angle(Dec, unit=u.deg)

    return DecAngle


def skyModelIdentify(origin, *args, **kwargs):
    """
    Identifies valid makesourcedb sky model files.
    """
    # Search for a format line. If found, assume file is valid
    try:
        if isinstance(args[0], basestring):
            f = open(args[0])
        elif isinstance(args[0], file_types):
            f = args[0]
        else:
            return False
        for line in f:
            if line.startswith("FORMAT") or line.startswith("format"):
                return True
        return False
    except UnicodeDecodeError:
        return False


def skyModelWriter(table, fileName):
    """
    Writes table to a makesourcedb sky model file.

    Parameters
    ----------
    fileName : str
        Output ASCII file to which the sky model is written

    """
    log = logging.getLogger('LSMTool.Write')

    modelFile = open(fileName, 'w')
    log.debug('Writing model to {0}'.format(fileName))

    # Make sure all columns have the correct makesourcedb units
    for colName in table.columns:
        units = allowedColumnUnits[colName.lower()]
        if units is not None:
            table[colName].convert_unit_to(units)

    # Add format line
    outLines = []
    formatString = []
    for colKey in table.keys():
        if colKey.lower() not in allowedColumnNames:
            continue
        colName = allowedColumnNames[colKey.lower()]

        if colName in table.meta and colName != 'Patch':
            colHeader = "{0}='{1}'".format(colName, table.meta[colName])
        elif colName == 'SpectralIndex':
            colHeader = "{0}='[]'".format(colName)
        else:
            colHeader = colName
        formatString.append(colHeader)
    outLines.append('FORMAT = {0}'.format(', '.join(formatString)))
    if 'History' in table.meta:
        outLines.append('\n\n# LSMTool history:\n# ')
        outLines.append('\n# '.join(table.meta['History']))
    outLines.append('\n')
    outLines.append('\n')

    # Add source lines
    if 'Patch' in table.keys():
        table = table.group_by('Patch')
        patchNames = table.groups.keys['Patch']
        for i, patchName in enumerate(patchNames):
            if patchName in table.meta:
                try:
                    gRA, gDec = table.meta[patchName]
                except ValueError:
                    raise ValueError('Multiple positions per patch. Please set'
                                     'the patch positions.')
            else:
                gRA = 0.0
                gDec = 0.0
            gRA = normalize_ra(gRA)
            gRAStr = Angle(gRA, unit='degree').to_string(unit='hourangle', sep=':', precision=4)
            gDec = normalize_dec(gDec)
            gDecStr = Angle(gDec, unit='degree').to_string(unit='degree', sep='.', precision=4)

            outLines.append(' , , {0}, {1}, {2}\n'.format(patchName, gRAStr,
                                                          gDecStr))
        for row in table.filled(fill_value=-9999):
            line = rowStr(row, table.meta)
            outLines.append(', '.join(line))
            outLines.append('\n')
    else:
        for row in table.filled(fill_value=-9999):
            line = rowStr(row, table.meta)
            outLines.append(', '.join(line))
            outLines.append('\n')

    modelFile.writelines(outLines)
    modelFile.close()


def rowStr(row, metaDict):
    """
    Returns makesourcedb representation of a row.

    Parameters
    ----------
    row : astropy.table.Row object
        Row to process
    metaDict : dict
        Table meta dictionary

    Returns
    -------
    line : str
        Sting representing a row in a makesourcedb sky model file

    """
    line = []
    for colKey in row.columns:
        try:
            colName = allowedColumnNames[colKey.lower()]
        except KeyError:
            continue

        # Determine whether the header (metaDict) defined a fill value and,
        # if so, use that for blank entries. If not, use the default value
        defaultVal = allowedColumnDefaults[colName.lower()]
        if colName in metaDict:
            fillVal = metaDict[colName]
            hasfillVal = True
        else:
            fillVal = defaultVal
            hasfillVal = False

        d = row[colKey]
        if str(d).startswith('-9999'):
            if hasfillVal:
                dstr = ' '
            else:
                dstr = str(fillVal)
        else:
            if type(d) is np.ndarray:
                if np.all(d == -9999):
                    if hasfillVal:
                        dstr = ' '
                    else:
                        dstr = str(fillVal)
                else:
                    dlist = d.tolist()
                    # Blank the value if it's equal to fill values
                    if hasfillVal and dlist == fillVal:
                        dlist = []
                    # Remove blanked values
                    if len(dlist) > 0:
                        while dlist[-1] == -9999:
                            dlist.pop()
                            if len(dlist) == 0:
                                break
                    dstr = str(dlist)
            else:
                if colKey == 'Ra':
                    d = normalize_ra(d)
                    dstr = Angle(d, unit='degree').to_string(unit='hourangle', sep=':')
                elif colKey == 'Dec':
                    d = normalize_dec(d)
                    dstr = Angle(d, unit='degree').to_string(unit='degree', sep='.')
                else:
                    dstr = str(d)
        line.append('{0}'.format(dstr))

    while line[-1] == ' ':
        line.pop()
    return line


def ds9RegionWriter(table, fileName):
    """
    Writes table to a ds9 region file.

    Parameters
    ----------
    table : astropy.table.Table object
        Input sky model table
    fileName : str
        Output file to which the sky model is written

    """
    log = logging.getLogger('LSMTool.Write')

    regionFile = open(fileName, 'w')
    log.debug('Writing ds9 region file to {0}'.format(fileName))

    outLines = []
    outLines.append('# Region file format: DS9 version 4.0\nglobal color=green '
                    'font="helvetica 10 normal" select=1 highlite=1 edit=1 '
                    'move=1 delete=1 include=1 fixed=0 source\nfk5\n')

    # Make sure all columns have the correct units
    for colName in table.columns:
        units = allowedColumnUnits[colName.lower()]
        if units is not None:
            table[colName].convert_unit_to(units)

    for row in table:
        ra = row['Ra']
        dec = row['Dec']
        name = row['Name']
        if row['Type'].lower() == 'gaussian':
            a = row['MajorAxis'] / 3600.0  # deg
            b = row['MinorAxis'] / 3600.0  # deg
            pa = row['Orientation']  # deg

            # ds9 can't handle 1-D Gaussians, so make sure they are 2-D
            if a < 1.0 / 3600.0:
                a = 1.0 / 3600.0  # deg
            if b < 1.0 / 3600.0:
                b = 1.0 / 3600.0  # deg
            region = 'ellipse({0}, {1}, {2}, {3}, {4}) # text={{{5}}}\n'.format(ra, dec, a, b,
                                                                                pa+90.0, name)
        else:
            region = 'point({0}, {1}) # point=cross width=2 text={{{2}}}\n'.format(ra, dec, name)
        outLines.append(region)

    regionFile.writelines(outLines)
    regionFile.close()


def kvisAnnWriter(table, fileName):
    """
    Writes table to a kvis annotation file.

    Parameters
    ----------
    table : astropy.table.Table object
        Input sky model table
    fileName : str
        Output file to which the sky model is written

    """
    log = logging.getLogger('LSMTool.Write')

    kvisFile = open(fileName, 'w')
    log.debug('Writing kvis annotation file to {0}'.format(fileName))

    # Make sure all columns have the correct units
    for colName in table.columns:
        units = allowedColumnUnits[colName.lower()]
        if units is not None:
            table[colName].convert_unit_to(units)

    outLines = []
    for row in table:
        ra = row['Ra']
        dec = row['Dec']
        name = row['Name']

        if row['Type'].lower() == 'gaussian':
            a = row['MajorAxis'] / 3600.0  # degree
            b = row['MinorAxis'] / 3600.0  # degree
            pa = row['Orientation']  # degree
            outLines.append('ELLIPSE W {0} {1} {2} {3} {4}\n'.format(ra, dec, a, b, pa))
        else:
            outLines.append('CIRCLE W {0} {1} 0.02\n'.format(ra, dec))
        outLines.append('TEXT W {0} {1} {2}\n'.format(ra - 0.07, dec, name))

    kvisFile.writelines(outLines)
    kvisFile.close()


def casaRegionWriter(table, fileName):
    """
    Writes model to a casa region file.

    Parameters
    ----------
    table : astropy.table.Table object
        Input sky model table
    fileName : str
        Output file to which the sky model is written

    """
    log = logging.getLogger('LSMTool.Write')

    casaFile = open(fileName, 'w')
    log.debug('Writing CASA box file to {0}'.format(fileName))

    outLines = []
    outLines.append('#CRTFv0\n')
    outLines.append('global coord=J2000\n\n')

    # Make sure all columns have the correct units
    for colName in table.columns:
        units = allowedColumnUnits[colName.lower()]
        if units is not None:
            table[colName].convert_unit_to(units)

    minSize = 10.0 / 3600.0  # min size in degrees
    for row in table:
        ra = row['Ra']
        dec = row['Dec']

        if row['Type'].lower() == 'gaussian':
            a = row['MajorAxis'] / 3600.0  # degree
            if a < minSize:
                a = minSize
            b = row['MinorAxis'] / 3600.0  # degree
            if b < minSize:
                b = minSize

            pa = row['Orientation']  # degree
            outLines.append('ellipse[[{0}deg, {1}deg], [{2}deg, {3}deg], '
                            '{4}deg]\n'.format(ra, dec, a, b, pa))
        else:
            outLines.append('ellipse[[{0}deg, {1}deg], [{2}deg, {3}deg], '
                            '{4}deg]\n'.format(ra, dec, minSize, minSize, 0.0))

    casaFile.writelines(outLines)
    casaFile.close()


def factorDirectionsWriter(table, fileName):
    """
    Writes patches to a Factor directions file.

    Note that Factor respects the order of patches and they are sorted here by
    apparent flux from brightest to faintest.

    Parameters
    ----------
    table : astropy.table.Table object
        Input sky model table; must have patches defined
    fileName : str
        Output file to which the sky model is written

    """
    log = logging.getLogger('LSMTool.Write')

    regionFile = open(fileName, 'w')
    log.debug('Writing Factor directions file to {0}'.format(fileName))

    outLines = []
    outLines.append('# name position atrous_do mscale_field_do cal_imsize '
                    'solint_ph solint_amp dynamic_range region_selfcal '
                    'region_facet peel_skymodel outlier_source cal_size_deg cal_flux_mJy\n')
    if 'History' in table.meta:
        outLines.append('\n# LSMTool history:\n# ')
        outLines.append('\n# '.join(table.meta['History']))
    outLines.append('\n')
    outLines.append('\n')

    # Make sure all columns have the correct units
    for colName in table.columns:
        units = allowedColumnUnits[colName.lower()]
        if units is not None:
            table[colName].convert_unit_to(units)

    table = table.group_by('Patch')
    patchNames = table.groups.keys['Patch']
    if 'patch_order' in table.meta:
        indx = table.meta['patch_order']
    else:
        indx = range(len(table.groups))
    if 'patch_size' in table.meta:
        sizes = table.meta['patch_size']
    else:
        sizes = [''] * len(table.groups)
    if 'patch_flux' in table.meta:
        fluxes = table.meta['patch_flux']
    else:
        fluxes = [''] * len(table.groups)
    for patchName, size, flux in zip(patchNames[indx], sizes[indx], fluxes[indx]):
        if patchName in table.meta:
            gRA, gDec = table.meta[patchName]
        else:
            gRA = Angle(0.0)
            gDec = Angle(0.0)
        outLines.append('{0} {1},{2} {3} {4} {5} {6} {7} {8} {9} {10} {11} {12} '
                        '{13} {14}\n'.format(patchName, gRA.to_string(unit='hourangle', sep='hms'),
                                             gDec.to_string(sep='dms'), 'empty', 'empty', 0, 0, 0,
                                             'LD', 'empty', 'empty', 'empty', False, size, flux))

    regionFile.writelines(outLines)
    regionFile.close()


def broadcastTable(fileName):
    """
    Sends a table via SAMP.

    Parameters
    ----------
    fileName : str
        Name of sky model file to broadcast

    """
    from astropy.vo.samp import SAMPIntegratedClient
    import urlparse

    client = SAMPIntegratedClient()
    client.connect()

    params = {}
    params["url"] = urlparse.urljoin('file:', os.path.abspath(fileName))
    params["name"] = "LSMTool sky model"
    message = {}
    message["samp.mtype"] = "table.load.votable"
    message["samp.params"] = params

    # Send message
    client.call_all('lsmtool', message)

    # Disconnect from the SAMP hub
    client.disconnect()


def coneSearch(VOService, position, radius):
    """
    Returns table from a VO cone search.

    Parameters
    ----------
    VOService : str
        Name of VO service to query (must be one of 'WENSS' or 'NVSS')
    position : list of floats
        A list specifying a position as [RA, Dec] in either makesourcedb
        format (e.g., ['12:23:43.21', '+22.34.21.2']) or in degrees (e.g.,
        [123.2312, 23.3422])
    radius : float or str, optional
        Radius in degrees (if float) or 'value unit' (if str; e.g.,
        '30 arcsec') for cone search region

    """
    import pyvo as vo

    log = logging.getLogger('LSMTool.Load')

    # Define allowed cone-search databases. These are the ones we know how to
    # convert to makesourcedb-formated sky models.
    columnMapping = {
        'nvss': {'NVSS': 'name', 'RAJ2000': 'ra', 'DEJ2000': 'dec', 'S1.4': 'i',
                 'MajAxis': 'majoraxis', 'MinAxis': 'minoraxis', 'referencefrequency': 1.4e9},
        'wenss': {'Name': 'name', 'RAJ2000': 'ra', 'DEJ2000': 'dec', 'Sint': 'i',
                  'MajAxis': 'majoraxis', 'MinAxis': 'minoraxis', 'PA': 'orientation',
                  'referencefrequency': 325e6}
        }

    if VOService.lower() in allowedVOServices:
        url = allowedVOServices[VOService.lower()]
    else:
        raise ValueError('VO query service not known. Allowed services are: '
                         '{0}'.format(allowedVOServices.keys()))

    # Get raw VO catalog
    log.debug('Querying VO service...')
    try:
        position = [RA2Angle(position[0])[0].value, Dec2Angle(position[1])[0].value]
    except TypeError:
        raise ValueError('VO query positon not understood.')
    try:
        radius = Angle(radius, unit='degree').value
    except TypeError:
        raise ValueError('VO query radius not understood.')
    VOcatalog = vo.conesearch(url, position, radius=radius)

    log.debug('Creating table...')
    try:
        table = Table.read(VOcatalog.votable)
    except IndexError:
        # Empty query result
        log.error('No sources found. Sky model is empty.')
        table = makeEmptyTable()
        return table
    table = convertExternalTable(table, columnMapping[VOService.lower()])

    return table


def convertExternalTable(table, columnMapping, fluxUnits='mJy'):
    """
    Converts an external table.

    Parameters
    ----------
    table : Table
        External table to convert
    columnMapping : dict
        Dict that defines the column name mapping from external table
    fluxUnits : str, optional
        Units of flux density in external table

    """
    log = logging.getLogger('LSMTool.Load')

    # Remove unneeded columns
    colsToRemove = []
    for colName in table.colnames:
        if colName not in columnMapping:
            colsToRemove.append(colName)
        elif columnMapping[colName] not in allowedColumnNames:
            colsToRemove.append(colName)
    for colName in colsToRemove:
        table.remove_column(colName)

    # Rename columns to match makesourcedb conventions
    for colName in table.colnames:
        if colName != allowedColumnNames[columnMapping[colName]]:
            table.rename_column(colName, allowedColumnNames[columnMapping[colName]])

    # Convert RA and Dec to Angle objects
    log.debug('Converting RA...')
    RARaw = table['Ra'].data.tolist()
    RACol = Column(name='Ra', data=RA2Angle(RARaw))
    RACol.format = raformat
    RAIndx = table.keys().index('Ra')
    table.remove_column('Ra')
    table.add_column(RACol, index=RAIndx)

    log.debug('Converting Dec...')
    DecRaw = table['Dec'].data.tolist()
    DecCol = Column(name='Dec', data=Dec2Angle(DecRaw))
    DecCol.format = decformat
    DecIndx = table.keys().index('Dec')
    table.remove_column('Dec')
    table.add_column(DecCol, index=DecIndx)

    # Make sure Name is a str column
    NameRaw = table['Name'].data.tolist()
    NameCol = Column(name='Name', data=NameRaw, dtype='{}100'.format(numpy_type))
    table.remove_column('Name')
    table.add_column(NameCol, index=0)

    # Convert flux and axis values to floats
    for name in ['I', 'MajorAxis', 'MinorAxis', 'Orientation']:
        if name in table.colnames:
            indx = table.index_column(name)
            intRaw = table[name].data.tolist()
            floatCol = Column(name=name, data=intRaw, dtype='float')
            table.remove_column(name)
            table.add_column(floatCol, index=indx)

    # Add source-type column
    types = ['POINT'] * len(table)
    if 'minoraxis' in columnMapping.values() and 'majoraxis' in columnMapping.values():
        for i, minor in enumerate(table[allowedColumnNames['minoraxis']]):
            if minor > 0.0:
                types[i] = 'GAUSSIAN'
            else:
                # Make sure semimajor axis and orientation are also 0 for POINT type
                table[allowedColumnNames['majoraxis']][i] = 0.0
                if 'orientation' in columnMapping.values():
                    table[allowedColumnNames['orientation']][i] = 0.0
    col = Column(name='Type', data=types, dtype='{}100'.format(numpy_type))
    table.add_column(col, index=1)

    # Add reference-frequency column
    refFreq = columnMapping['referencefrequency']
    col = Column(name='ReferenceFrequency', data=np.array([refFreq]*len(table), dtype=float))
    table.add_column(col)

    # Set column units and default values
    for i, colName in enumerate(table.colnames):
        log.debug("Setting units for column '{0}' to {1}".format(
            colName, allowedColumnUnits[colName.lower()]))
        if colName == 'I':
            table.columns[colName].unit = fluxUnits
            table.columns[colName].convert_unit_to('Jy')
            table.columns[colName].format = fluxformat
        else:
            table.columns[colName].unit = allowedColumnUnits[colName.lower()]

        if hasattr(table.columns[colName], 'filled') and allowedColumnDefaults[colName.lower()] is not None:
            # Note: we used deepcopy() here to ensure that the original
            # is not altered by later changes to fillVal
            fillVal = deepcopy(allowedColumnDefaults)[colName.lower()]
            log.debug("Setting default value for column '{0}' to {1}".
                      format(colName, fillVal))
            if colName == 'SpectralIndex':
                # We cannot set the fill value to a list/array, so just use a float
                fillVal = 0.0
            table.columns[colName].set_fill_value(fillVal)

    return table


def getTGSS(position, radius):
    """
    Returns the file name from a TGSS search.

    Parameters
    ----------
    position : list of floats
        A list specifying a position as [RA, Dec] in either makesourcedb
        format (e.g., ['12:23:43.21', '+22.34.21.2']) or in degrees (e.g.,
        [123.2312, 23.3422])
    radius : float or str, optional
        Radius in degrees (if float) or 'value unit' (if str; e.g.,
        '30 arcsec') for cone search region

    """
    import tempfile
    import subprocess

    log = logging.getLogger('LSMTool.Load')

    outFile = tempfile.NamedTemporaryFile()
    RA = RA2Angle(position[0])[0].value
    Dec = Dec2Angle(position[1])[0].value
    try:
        radius = Angle(radius, unit='degree').value
    except TypeError:
        raise ValueError('TGSS query radius "{}" not understood.'.format(radius))

    log.debug('Querying TGSS...')
    url = TGSS_URL + '?coord={0},{1}&radius={2}&unit=deg&deconv=y'.format(RA, Dec, radius)
    cmd = ['wget', '-O', outFile.name, url]
    subprocess.call(cmd)

    return outFile


def getGSM(position, radius):
    """
    Returns the file name from a GSM search.

    Parameters
    ----------
    position : list of floats
        A list specifying a position as [RA, Dec] in either makesourcedb
        format (e.g., ['12:23:43.21', '+22.34.21.2']) or in degrees (e.g.,
        [123.2312, 23.3422])
    radius : float or str, optional
        Radius in degrees (if float) or 'value unit' (if str; e.g.,
        '30 arcsec') for cone search region

    """
    import tempfile
    import subprocess

    log = logging.getLogger('LSMTool.Load')

    outFile = tempfile.NamedTemporaryFile()
    RA = RA2Angle(position[0])[0].value
    Dec = Dec2Angle(position[1])[0].value
    try:
        radius = Angle(radius, unit='degree').value
    except TypeError:
        raise ValueError('GSM query radius "{}" not understood.'.format(radius))

    log.debug('Querying GSM...')
    url = GSM_URL + '?coord={0},{1}&radius={2}&unit=deg&deconv=y'.format(RA, Dec, radius)
    cmd = ['wget', '-O', outFile.name, url]
    subprocess.call(cmd)

    return outFile


def getLoTSS(position, radius):
    """
    Returns table from a LoTSS search.

    Parameters
    ----------
    position : list of floats
        A list specifying a position as [RA, Dec] in either makesourcedb
        format (e.g., ['12:23:43.21', '+22.34.21.2']) or in degrees (e.g.,
        [123.2312, 23.3422])
    radius : float or str, optional
        Radius in degrees (if float) or 'value unit' (if str; e.g.,
        '30 arcsec') for cone search region

    """
    import tempfile
    import subprocess

    log = logging.getLogger('LSMTool.Load')

    columnMapping = {'Source_Name': 'name', 'RA': 'ra', 'DEC': 'dec', 'Total_flux': 'i',
                     'DC_Maj': 'majoraxis', 'DC_Min': 'minoraxis', 'PA': 'orientation',
                     'referencefrequency': 1.4e8}

    outFile = tempfile.NamedTemporaryFile()
    RA = RA2Angle(position[0])[0].value
    Dec = Dec2Angle(position[1])[0].value
    try:
        radius = Angle(radius, unit='degree').value * 60.0  # arcmin
    except TypeError:
        raise ValueError('LoTSS query radius "{}" not understood.'.format(radius))

    log.debug('Querying LoTSS...')
    url = (LOTSS_URL + '?__nevow_form__=genForm&'
           'hscs_pos={0}%2C%20{1}&hscs_sr={2}&_DBOPTIONS_ORDER=&'
           '_DBOPTIONS_DIR=ASC&MAXREC=100000&_FORMAT=CSV&submit=Go'.format(RA, Dec, radius))
    cmd = ['wget', '-O', outFile.name, url]
    subprocess.call(cmd)

    # Read in the table
    table = Table.read(outFile.name, format='ascii.csv', header_start=0)
    table = convertExternalTable(table, columnMapping)

    return table


def makeEmptyTable():
    """
    Returns an empty sky model table.
    """
    outlines = ['Z, Z, 0.0, 0.0, 0.0\n']
    colNames = ['Name', 'Type', 'Ra', 'Dec', 'I']
    converters = {}
    nameCol = 'col{0}'.format(colNames.index('Name')+1)
    converters[nameCol] = [ascii.convert_numpy('{}100'.format(numpy_type))]
    typeCol = 'col{0}'.format(colNames.index('Type')+1)
    converters[typeCol] = [ascii.convert_numpy('{}100'.format(numpy_type))]
    table = Table.read(outlines, guess=False, format='ascii.no_header', delimiter=',',
                       names=colNames, comment='#', data_start=0, converters=converters)
    table.remove_rows(0)
    return table


# Register the file reader, identifier, and writer functions with astropy.io
registry.register_reader('makesourcedb', Table, skyModelReader)
registry.register_identifier('makesourcedb', Table, skyModelIdentify)
registry.register_writer('makesourcedb', Table, skyModelWriter)
registry.register_writer('ds9', Table, ds9RegionWriter)
registry.register_writer('kvis', Table, kvisAnnWriter)
registry.register_writer('casa', Table, casaRegionWriter)
registry.register_writer('factor', Table, factorDirectionsWriter)
